from __future__ import annotations

import datetime as dt
import sys
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, SupportsIndex, overload

import hightime as ht
import numpy as np
import numpy.typing as npt
from typing_extensions import Self

from nitypes._arguments import arg_to_uint, validate_dtype, validate_unsupported_arg
from nitypes._exceptions import invalid_arg_type, invalid_array_ndim
from nitypes._numpy import asarray as _np_asarray
from nitypes.time.typing import AnyDateTime, AnyTimeDelta
from nitypes.waveform._digital._port import bit_mask, get_port_dtype, port_to_line_data
from nitypes.waveform._exceptions import (
    create_capacity_mismatch_error,
    create_capacity_too_small_error,
    create_datatype_mismatch_error,
    create_irregular_timestamp_count_mismatch_error,
    create_signal_count_mismatch_error,
    create_start_index_or_sample_count_too_large_error,
    create_start_index_too_large_error,
)
from nitypes.waveform._extended_properties import CHANNEL_NAME, LINE_NAMES
from nitypes.waveform._types import DIGITAL_PORT_DTYPES, DIGITAL_STATE_DTYPES
from nitypes.waveform.typing import (
    ExtendedPropertyValue,
    TDigitalState,
    TOtherDigitalState,
)

if sys.version_info < (3, 10):
    import array as std_array

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import (
        DigitalState,
        DigitalWaveformSignalCollection,
        ExtendedPropertyDictionary,
        Timing,
    )
else:
    from nitypes.waveform._digital._signal_collection import (
        DigitalWaveformSignalCollection,
    )
    from nitypes.waveform._digital._state import DigitalState
    from nitypes.waveform._extended_properties import ExtendedPropertyDictionary
    from nitypes.waveform._timing import Timing


@dataclass(frozen=True)
class DigitalWaveformFailure:
    """A test failure, indicating where the actual waveform did not match the expected waveform."""

    sample_index: int
    """The sample index into the compared waveform where the test failure occurred."""

    expected_sample_index: int
    """The sample index into the expected waveform where the test failure occurred."""

    signal_index: int
    """The signal index where the test failure occurred."""

    actual_state: DigitalState
    """The state from the compared waveform where the test failure occurred."""

    expected_state: DigitalState
    """The state from the expected waveform where the test failure occurred."""


@dataclass(frozen=True)
class DigitalWaveformTestResult:
    """A test result from comparing a digital waveform against an expected digital waveform."""

    @property
    def success(self) -> bool:
        """True if the test is successful, False if the test failed."""
        return len(self.failures) == 0

    failures: Sequence[DigitalWaveformFailure]
    """A collection of test failure information."""


class DigitalWaveform(Generic[TDigitalState]):
    """A digital waveform, which encapsulates digital data and timing information.

    Constructing
    ^^^^^^^^^^^^

    To construct a digital waveform, use the :class:`DigitalWaveform` class:

    >>> DigitalWaveform()
    nitypes.waveform.DigitalWaveform(0, 1)
    >>> DigitalWaveform(sample_count=5, signal_count=3)  # doctest: +NORMALIZE_WHITESPACE
    nitypes.waveform.DigitalWaveform(5, 3, data=array([[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0],
    [0, 0, 0]], dtype=uint8))

    When displaying a digital waveform as a string, the first number is the sample count and the second
    number is the signal count.

    To construct a digital waveform from a NumPy array of line data, use the
    :any:`DigitalWaveform.from_lines` method. Each array element represents a digital state, such as 1
    for "on" or 0 for "off". The line data should be in a 1D array indexed by sample or a 2D array
    indexed by (sample, signal). The digital waveform displays the line data as a 2D array.

    >>> import numpy as np
    >>> DigitalWaveform.from_lines(np.array([0, 1, 0], np.uint8))
    nitypes.waveform.DigitalWaveform(3, 1, data=array([[0], [1], [0]], dtype=uint8))
    >>> DigitalWaveform.from_lines(np.array([[0, 0], [1, 0], [0, 1], [1, 1]], np.uint8))
    nitypes.waveform.DigitalWaveform(4, 2, data=array([[0, 0], [1, 0], [0, 1], [1, 1]], dtype=uint8))

    You can also use :any:`DigitalWaveform.from_lines` to construct a digital waveform from a sequence,
    such as a list.

    >>> DigitalWaveform.from_lines([[0, 0], [1, 0], [0, 1], [1, 1]])
    nitypes.waveform.DigitalWaveform(4, 2, data=array([[0, 0], [1, 0], [0, 1], [1, 1]], dtype=uint8))

    To construct a digital waveform from a NumPy array of port data, use the
    :any:`DigitalWaveform.from_port` method. Each element of the port data array represents a digital
    sample taken over a port of signals. Each bit in the sample is a signal value, either 1 for "on" or
    0 for "off". The least significant bit of the sample is placed at signal index 0 of the
    DigitalWaveform.

    >>> DigitalWaveform.from_port(np.array([0, 1, 2, 3], np.uint8))  # doctest: +NORMALIZE_WHITESPACE
    nitypes.waveform.DigitalWaveform(4, 8, data=array([[0, 0, 0, 0, 0, 0, 0, 0],
    [1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0], [1, 1, 0, 0, 0, 0, 0, 0]], dtype=uint8))

    You can use a mask to specify which lines in the port to include in the waveform.

    >>> DigitalWaveform.from_port(np.array([0, 1, 2, 3], np.uint8), 0x3)
    nitypes.waveform.DigitalWaveform(4, 2, data=array([[0, 0], [1, 0], [0, 1], [1, 1]], dtype=uint8))

    You can also use a non-NumPy sequence such as a list, but you must specify a mask so the waveform
    knows how many bits are in each list element.

    >>> DigitalWaveform.from_port([0, 1, 2, 3], 0x3)
    nitypes.waveform.DigitalWaveform(4, 2, data=array([[0, 0], [1, 0], [0, 1], [1, 1]], dtype=uint8))

    The 2D version, :any:`DigitalWaveform.from_ports`, returns multiple waveforms, one for each row of
    data in the array or nested sequence.

    >>> nested_list = [[0, 1, 2, 3], [3, 0, 3, 0]]
    >>> DigitalWaveform.from_ports(nested_list, [0x3, 0x3])  # doctest: +NORMALIZE_WHITESPACE
    [nitypes.waveform.DigitalWaveform(4, 2, data=array([[0, 0], [1, 0], [0, 1], [1, 1]], dtype=uint8)),
    nitypes.waveform.DigitalWaveform(4, 2, data=array([[1, 1], [0, 0], [1, 1], [0, 0]], dtype=uint8))]

    Digital signals
    ^^^^^^^^^^^^^^^

    You can access individual signals using the :any:`DigitalWaveform.signals` property.

    >>> wfm = DigitalWaveform.from_port([0, 1, 2, 3], 0x3)
    >>> wfm.signals[0]
    nitypes.waveform.DigitalWaveformSignal(data=array([0, 1, 0, 1], dtype=uint8))
    >>> wfm.signals[1]
    nitypes.waveform.DigitalWaveformSignal(data=array([0, 0, 1, 1], dtype=uint8))

    The :any:`DigitalWaveformSignal.data` property returns a view of the data for that signal.

    >>> wfm.signals[0].data
    array([0, 1, 0, 1], dtype=uint8)

    Digital signal names
    ^^^^^^^^^^^^^^^^^^^^

    The :any:`DigitalWaveformSignal.name` property allows you to get and set the signal names.

    >>> wfm.signals[0].name = "port0/line0"
    >>> wfm.signals[1].name = "port0/line1"
    >>> wfm.signals[0].name
    'port0/line0'
    >>> wfm.signals[0]
    nitypes.waveform.DigitalWaveformSignal(name='port0/line0', data=array([0, 1, 0, 1], dtype=uint8))

    The signal names are stored in the ``NI_LineNames`` extended property on the digital waveform.

    >>> wfm.extended_properties["NI_LineNames"]
    'port0/line0, port0/line1'

    When creating a digital waveform, you can directly set the ``NI_LineNames`` extended property.

    >>> wfm = DigitalWaveform.from_port([2, 4], 0x7,
    ... extended_properties={"NI_LineNames": "Dev1/port1/line4, Dev1/port1/line5, Dev1/port1/line6"})
    >>> wfm.signals[0]
    nitypes.waveform.DigitalWaveformSignal(name='Dev1/port1/line4', data=array([0, 0], dtype=uint8))
    >>> wfm.signals[1]
    nitypes.waveform.DigitalWaveformSignal(name='Dev1/port1/line5', data=array([1, 0], dtype=uint8))
    >>> wfm.signals[2]
    nitypes.waveform.DigitalWaveformSignal(name='Dev1/port1/line6', data=array([0, 1], dtype=uint8))

    Digital state types
    ^^^^^^^^^^^^^^^^^^^

    By default, digital waveforms use a NumPy ``dtype`` of :any:`numpy.uint8`, which uses a byte of
    memory for each digital state.

    Using ``np.uint8`` allows the waveform to contain digital states other than "on" or off", such as
    such as :any:`DigitalState.FORCE_OFF` (``X``) or :any:`DigitalState.COMPARE_HIGH` (``H``). This
    capability is used for digital pattern applications.

    You can also construct a digital waveform using a NumPy ``dtype`` of :any:`numpy.bool`. This also
    uses a byte of memory for each digital state, but it restricts the states to "on" and "off".

    Testing digital waveforms
    ^^^^^^^^^^^^^^^^^^^^^^^^^

    You can use :meth:`DigitalWaveform.test` to compare an acquired waveform against an expected
    waveform. This returns a :class:`DigitalWaveformTestResult` object, which has a Boolean ``success``
    property and a ``failures`` property containing a collection of :class:`DigitalWaveformFailure`
    objects, which indicate the location of each test failure.

    Here is an example. The expected waveform counts in binary using ``COMPARE_LOW`` (``L``) and
    ``COMPARE_HIGH`` (``H``), but signal 1 of the actual waveform is stuck high.

    >>> actual = DigitalWaveform.from_lines([[0, 1], [1, 1], [0, 1], [1, 1]])
    >>> expected = DigitalWaveform.from_lines([[DigitalState.COMPARE_LOW, DigitalState.COMPARE_LOW],
    ... [DigitalState.COMPARE_HIGH, DigitalState.COMPARE_LOW],
    ... [DigitalState.COMPARE_LOW, DigitalState.COMPARE_HIGH],
    ... [DigitalState.COMPARE_HIGH, DigitalState.COMPARE_HIGH]])
    >>> result = actual.test(expected)
    >>> result.success
    False
    >>> len(result.failures)
    2

    The failures indicate the sample indices into the actual and expected waveforms, the signal index,
    and the digital state from the actual and expected waveforms:

    >>> result.failures[0]  # doctest: +NORMALIZE_WHITESPACE
    DigitalWaveformFailure(sample_index=0, expected_sample_index=0, signal_index=1,
    actual_state=<DigitalState.FORCE_UP: 1>, expected_state=<DigitalState.COMPARE_LOW: 3>)
    >>> result.failures[1]  # doctest: +NORMALIZE_WHITESPACE
    DigitalWaveformFailure(sample_index=1, expected_sample_index=1, signal_index=1,
    actual_state=<DigitalState.FORCE_UP: 1>, expected_state=<DigitalState.COMPARE_LOW: 3>)

    Timing information
    ^^^^^^^^^^^^^^^^^^

    Digital waveforms have the same timing information as analog waveforms.

    Class members
    ^^^^^^^^^^^^^
    """  # noqa: W505 - doc line too long

    @overload
    @classmethod
    def from_lines(
        cls,
        array: npt.NDArray[TOtherDigitalState],
        dtype: None = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        signal_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> DigitalWaveform[TOtherDigitalState]: ...

    @overload
    @classmethod
    def from_lines(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: type[TOtherDigitalState] | np.dtype[TOtherDigitalState],
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        signal_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> DigitalWaveform[TOtherDigitalState]: ...

    @overload
    @classmethod
    def from_lines(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: npt.DTypeLike = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        signal_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> DigitalWaveform[Any]: ...

    @classmethod
    def from_lines(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: npt.DTypeLike = None,
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        signal_count: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
    ) -> DigitalWaveform[Any]:
        """Construct a waveform from a one or two-dimensional array or sequence of line data.

        Each element of the line data array represents a digital state, such as 1 for "on" or 0
        for "off". The line data should be in a 1D array indexed by sample or a 2D array indexed
        by (sample, signal). The line data may also use digital state values from the
        :class:`DigitalState` enum.

        Args:
            array: The line data as a one or two-dimensional array or a sequence.
            dtype: The NumPy data type for the waveform data.
            copy: Specifies whether to copy the array or save a reference to it.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
            signal_count: The number of signals in the waveform.
            extended_properties: The extended properties of the waveform.
            timing: The timing information of the waveform.

        Returns:
            A waveform containing the specified data.
        """
        if isinstance(array, np.ndarray):
            if array.ndim not in (1, 2):
                raise invalid_array_ndim(
                    "input array", "one or two-dimensional array or sequence", array.ndim
                )
            if dtype is not None and array.dtype != dtype:
                raise create_datatype_mismatch_error("input array", array.dtype, "requested", dtype)
        elif isinstance(array, Sequence) or (
            sys.version_info < (3, 10) and isinstance(array, std_array.array)
        ):
            if dtype is None:
                dtype = np.uint8
        else:
            raise invalid_arg_type("input array", "one or two-dimensional array or sequence", array)

        return cls(
            data=_np_asarray(array, dtype, copy=copy),
            start_index=start_index,
            sample_count=sample_count,
            signal_count=signal_count,
            extended_properties=extended_properties,
            timing=timing,
        )

    @overload
    @classmethod
    def from_port(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        mask: SupportsIndex | None = ...,
        dtype: None = ...,
        *,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> DigitalWaveform[np.uint8]: ...

    @overload
    @classmethod
    def from_port(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        mask: SupportsIndex | None = ...,
        dtype: (
            type[TOtherDigitalState]  # pyright: ignore[reportInvalidTypeVarUse]
            | np.dtype[TOtherDigitalState]
        ) = ...,
        *,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> DigitalWaveform[TOtherDigitalState]: ...

    @overload
    @classmethod
    def from_port(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        mask: SupportsIndex | None = ...,
        dtype: npt.DTypeLike = ...,
        *,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> DigitalWaveform[Any]: ...

    @classmethod
    def from_port(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        mask: SupportsIndex | None = None,
        dtype: npt.DTypeLike = None,
        *,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
    ) -> DigitalWaveform[Any]:
        """Construct a waveform from a one-dimensional array or sequence of port data.

        This method allocates a new array in order to convert the port data to line data.

        Each element of the port data array represents a digital sample taken over a port of
        signals. Each bit in the sample represents a digital state, either 1 for "on" or 0 for
        "off". The least significant bit of the sample is placed at signal index 0 of the
        DigitalWaveform.

        If the input array is not a NumPy array, you must specify the mask.

        Args:
            array: The port data as a one-dimensional array or a sequence.
            mask: A bitmask specifying which lines to include in the waveform.
            dtype: The NumPy data type for the waveform (line) data.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
            extended_properties: The extended properties of the waveform.
            timing: The timing information of the waveform.

        Returns:
            A waveform containing the specified data.
        """
        if isinstance(array, np.ndarray):
            if array.ndim != 1:
                raise invalid_array_ndim(
                    "input array", "one-dimensional array or sequence", array.ndim
                )
            validate_dtype(array.dtype, DIGITAL_PORT_DTYPES)
            default_mask = bit_mask(array.dtype.itemsize * 8)
        elif isinstance(array, Sequence) or (
            sys.version_info < (3, 10) and isinstance(array, std_array.array)
        ):
            # np.array([0,1]).dtype is np.int64 by default, so the user must specify the port size.
            if mask is None:
                raise ValueError(
                    "You must specify a mask when the input array is not a NumPy array."
                )
            default_mask = 0
        else:
            raise invalid_arg_type("input array", "one or two-dimensional array or sequence", array)

        if dtype is None:
            dtype = np.uint8
        validate_dtype(dtype, DIGITAL_STATE_DTYPES)

        mask = arg_to_uint("mask", mask, default_mask)

        if isinstance(array, np.ndarray):
            port_dtype = array.dtype
        else:
            port_dtype = get_port_dtype(mask)

        port_data = _np_asarray(array, port_dtype)
        line_data = port_to_line_data(port_data, mask)
        if line_data.dtype != dtype:
            line_data = line_data.view(dtype)

        return cls(
            data=line_data,
            dtype=dtype,
            start_index=start_index,
            sample_count=sample_count,
            extended_properties=extended_properties,
            timing=timing,
        )

    @overload
    @classmethod
    def from_ports(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        masks: Sequence[SupportsIndex] | None = ...,
        dtype: None = ...,
        *,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> Sequence[DigitalWaveform[np.uint8]]: ...

    @overload
    @classmethod
    def from_ports(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        masks: Sequence[SupportsIndex] | None = ...,
        dtype: (
            type[TOtherDigitalState]  # pyright: ignore[reportInvalidTypeVarUse]
            | np.dtype[TOtherDigitalState]
        ) = ...,
        *,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> Sequence[DigitalWaveform[TOtherDigitalState]]: ...

    @overload
    @classmethod
    def from_ports(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        masks: Sequence[SupportsIndex] | None = ...,
        dtype: npt.DTypeLike = ...,
        *,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> Sequence[DigitalWaveform[Any]]: ...

    @classmethod
    def from_ports(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        masks: Sequence[SupportsIndex] | None = None,
        dtype: npt.DTypeLike = None,
        *,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
    ) -> Sequence[DigitalWaveform[Any]]:
        """Construct a waveform from a two-dimensional array or sequence of port data.

        This method allocates a new array in order to convert the port data to line data.

        Each row of the port data array corresponds to a resulting DigitalWaveform. Each element of
        the port data array represents a digital sample taken over a port of signals. Each bit in
        the sample is represents a digital state, either 1 for "on" or 0 for "off". The least
        significant bit of the sample is placed at signal index 0 of the corresponding
        DigitalWaveform.

        If the input array is not a NumPy array, you must specify the masks.

        Args:
            array: The port data as a two-dimensional array or a sequence.
            masks: A sequence of bitmasks specifying which lines from each port to include in the
                corresponding waveform.
            dtype: The NumPy data type for the waveform (line) data.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
            extended_properties: The extended properties of the waveform.
            timing: The timing information of the waveform.

        Returns:
            A waveform containing the specified data.
        """
        if isinstance(array, np.ndarray):
            if array.ndim != 2:
                raise invalid_array_ndim(
                    "input array", "two-dimensional array or sequence", array.ndim
                )
            validate_dtype(array.dtype, DIGITAL_PORT_DTYPES)
            default_masks = [bit_mask(array.dtype.itemsize * 8)] * array.shape[0]
        elif isinstance(array, Sequence) or (
            sys.version_info < (3, 10) and isinstance(array, std_array.array)
        ):
            # np.array([0,1]).dtype is np.int64 by default, so the user must specify the port size.
            if masks is None:
                raise ValueError(
                    "You must specify the masks when the input array is not a NumPy array."
                )
            default_masks = []
        else:
            raise invalid_arg_type("input array", "one or two-dimensional array or sequence", array)

        if dtype is None:
            dtype = np.uint8
        validate_dtype(dtype, DIGITAL_STATE_DTYPES)

        if not isinstance(masks, (type(None), Sequence)):
            raise invalid_arg_type("masks", "sequence of ints")
        if masks is not None:
            validated_masks = [arg_to_uint("mask", mask) for mask in masks]
        else:
            validated_masks = default_masks

        if isinstance(array, np.ndarray):
            port_dtype = array.dtype
        else:
            port_dtype = get_port_dtype(validated_masks)

        port_data = _np_asarray(array, port_dtype)
        waveforms = []
        for port_index in range(port_data.shape[0]):
            line_data = port_to_line_data(port_data[port_index, :], validated_masks[port_index])
            if line_data.dtype != dtype:
                line_data = line_data.view(dtype)

            waveforms.append(
                cls(
                    data=line_data,
                    dtype=dtype,
                    start_index=start_index,
                    sample_count=sample_count,
                    extended_properties=extended_properties,
                    timing=timing,
                )
            )
        return waveforms

    __slots__ = [
        "_data",
        "_data_1d",
        "_start_index",
        "_sample_count",
        "_extended_properties",
        "_timing",
        "_signals",
        "_signal_names",
        "__weakref__",
    ]

    _data: npt.NDArray[TDigitalState]
    _data_1d: npt.NDArray[TDigitalState] | None
    _start_index: int
    _sample_count: int
    _extended_properties: ExtendedPropertyDictionary
    _timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta]
    _signals: DigitalWaveformSignalCollection[TDigitalState] | None
    _signal_names: list[str] | None

    # If neither dtype nor data is specified, _TData defaults to np.uint8.
    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: DigitalWaveform[np.uint8],
        sample_count: SupportsIndex | None = ...,
        signal_count: SupportsIndex | None = ...,
        dtype: None = ...,
        default_value: bool | int | DigitalState | None = ...,
        *,
        data: None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: DigitalWaveform[TOtherDigitalState],
        sample_count: SupportsIndex | None = ...,
        signal_count: SupportsIndex | None = ...,
        dtype: type[TOtherDigitalState] | np.dtype[TOtherDigitalState] = ...,
        default_value: bool | int | DigitalState | None = ...,
        *,
        data: None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: DigitalWaveform[TOtherDigitalState],
        sample_count: SupportsIndex | None = ...,
        signal_count: SupportsIndex | None = ...,
        dtype: None = ...,
        default_value: bool | int | DigitalState | None = ...,
        *,
        data: npt.NDArray[TOtherDigitalState] = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: DigitalWaveform[Any],
        sample_count: SupportsIndex | None = ...,
        signal_count: SupportsIndex | None = ...,
        dtype: npt.DTypeLike = ...,
        default_value: bool | int | DigitalState | None = ...,
        *,
        data: npt.NDArray[Any] | None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = ...,
    ) -> None: ...

    def __init__(
        self,
        sample_count: SupportsIndex | None = None,
        signal_count: SupportsIndex | None = None,
        dtype: npt.DTypeLike = None,
        default_value: bool | int | DigitalState | None = None,
        *,
        data: npt.NDArray[Any] | None = None,
        start_index: SupportsIndex | None = None,
        capacity: SupportsIndex | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        copy_extended_properties: bool = True,
        timing: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta] | None = None,
    ) -> None:
        """Initialize a new digital waveform.

        Args:
            sample_count: The number of samples in the waveform.
            signal_count: The number of signals in the waveform.
            dtype: The NumPy data type for the waveform data.
            default_value: The :class:`DigitalState` to initialize the waveform with.
            data: A NumPy ndarray to use for sample storage. The waveform takes ownership
                of this array. If not specified, an ndarray is created based on the specified dtype,
                start index, sample count, and capacity.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
            capacity: The number of samples to allocate. Pre-allocating a larger buffer optimizes
                appending samples to the waveform.
            extended_properties: The extended properties of the waveform.
            copy_extended_properties: Specifies whether to copy the extended properties or take
                ownership.
            timing: The timing information of the waveform.

        Returns:
            A digital waveform.
        """
        if data is None:
            self._init_with_new_array(
                sample_count,
                signal_count,
                dtype,
                default_value,
                start_index=start_index,
                capacity=capacity,
            )
        elif isinstance(data, np.ndarray):
            self._init_with_provided_array(
                data,
                dtype,
                start_index=start_index,
                sample_count=sample_count,
                signal_count=signal_count,
                capacity=capacity,
            )
        else:
            raise invalid_arg_type("raw data", "NumPy ndarray", data)

        if copy_extended_properties or not isinstance(
            extended_properties, ExtendedPropertyDictionary
        ):
            extended_properties = ExtendedPropertyDictionary(extended_properties)
        self._extended_properties = extended_properties

        if timing is None:
            timing = Timing.empty
        self._timing = timing

        self._signals = None
        self._signal_names = None

    def _init_with_new_array(
        self,
        sample_count: SupportsIndex | None = None,
        signal_count: SupportsIndex | None = None,
        dtype: npt.DTypeLike = None,
        default_value: bool | int | DigitalState | None = None,
        *,
        start_index: SupportsIndex | None = None,
        capacity: SupportsIndex | None = None,
    ) -> None:
        start_index = arg_to_uint("start index", start_index, 0)
        sample_count = arg_to_uint("sample count", sample_count, 0)
        signal_count = arg_to_uint("signal count", signal_count, 1)
        capacity = arg_to_uint("capacity", capacity, sample_count)

        if dtype is None:
            dtype = np.uint8
        validate_dtype(dtype, DIGITAL_STATE_DTYPES)

        if start_index > capacity:
            raise create_start_index_too_large_error(start_index, "capacity", capacity)
        if start_index + sample_count > capacity:
            raise create_start_index_or_sample_count_too_large_error(
                start_index, sample_count, "capacity", capacity
            )

        if default_value is None:
            default_value = 0
        elif not isinstance(default_value, (bool, int, DigitalState)):
            raise invalid_arg_type("default value", "bool, int, or DigitalState", default_value)

        self._data = np.full((capacity, signal_count), default_value, dtype)
        self._data_1d = None
        self._start_index = start_index
        self._sample_count = sample_count

    def _init_with_provided_array(
        self,
        data: npt.NDArray[TDigitalState],
        dtype: npt.DTypeLike = None,
        *,
        start_index: SupportsIndex | None = None,
        sample_count: SupportsIndex | None = None,
        signal_count: SupportsIndex | None = None,
        capacity: SupportsIndex | None = None,
    ) -> None:
        if not isinstance(data, np.ndarray):
            raise invalid_arg_type("input array", "one or two-dimensional array", data)

        if dtype is None:
            dtype = data.dtype
        if dtype != data.dtype:
            raise create_datatype_mismatch_error(
                "input array", data.dtype, "requested", np.dtype(dtype)
            )
        validate_dtype(dtype, DIGITAL_STATE_DTYPES)

        if data.ndim == 1:
            data_signal_count = 1
            data_1d = data
            data = data.reshape(len(data), 1)
        elif data.ndim == 2:
            data_signal_count = data.shape[1]
            data_1d = None
        else:
            raise invalid_array_ndim("input array", "one or two-dimensional array", data.ndim)

        capacity = arg_to_uint("capacity", capacity, len(data))
        if capacity != len(data):
            raise create_capacity_mismatch_error(capacity, len(data))

        start_index = arg_to_uint("start index", start_index, 0)
        if start_index > capacity:
            raise create_start_index_too_large_error(start_index, "input array length", capacity)

        sample_count = arg_to_uint("sample count", sample_count, len(data) - start_index)
        if start_index + sample_count > len(data):
            raise create_start_index_or_sample_count_too_large_error(
                start_index, sample_count, "input array length", len(data)
            )

        signal_count = arg_to_uint("signal count", signal_count, data_signal_count)
        if signal_count != data_signal_count:
            raise create_signal_count_mismatch_error(
                "provided", signal_count, "array", data_signal_count
            )

        self._data = data
        self._data_1d = data_1d
        self._start_index = start_index
        self._sample_count = sample_count

    @property
    def signals(self) -> DigitalWaveformSignalCollection[TDigitalState]:
        """A collection of objects representing waveform signals."""
        # Lazily allocate self._signals if the application needs it.
        #
        # https://github.com/ni/nitypes-python/issues/131 - DigitalWaveform.signals introduces a
        # reference cycle, which affects GC overhead.
        value = self._signals
        if value is None:
            value = self._signals = DigitalWaveformSignalCollection(self)
        return value

    @property
    def data(self) -> npt.NDArray[TDigitalState]:
        """The waveform data, indexed by (sample, signal)."""
        return self._data[self._start_index : self._start_index + self._sample_count]

    def get_data(
        self, start_index: SupportsIndex | None = 0, sample_count: SupportsIndex | None = None
    ) -> npt.NDArray[TDigitalState]:
        """Get a subset of the waveform data.

        Args:
            start_index: The sample index at which the data begins.
            sample_count: The number of samples to return.

        Returns:
            A subset of the raw waveform data.
        """
        start_index = arg_to_uint("start index", start_index, 0)
        if start_index > self.sample_count:
            raise create_start_index_too_large_error(
                start_index, "number of samples in the waveform", self.sample_count
            )

        sample_count = arg_to_uint("sample count", sample_count, self.sample_count - start_index)
        if start_index + sample_count > self.sample_count:
            raise create_start_index_or_sample_count_too_large_error(
                start_index, sample_count, "number of samples in the waveform", self.sample_count
            )

        return self.data[start_index : start_index + sample_count]

    @property
    def sample_count(self) -> int:
        """The number of samples in the waveform."""
        return self._sample_count

    @property
    def signal_count(self) -> int:
        """The number of signals in the waveform."""
        # npt.NDArray[_ScalarT] currently has a shape type of _AnyShape, which is tuple[Any, ...]
        shape: tuple[int, ...] = self._data.shape
        return shape[1]

    @property
    def capacity(self) -> int:
        """The total capacity available for waveform data.

        Setting the capacity resizes the underlying NumPy array in-place.

        * Other Python objects with references to the array will see the array size change.
        * If the array has a reference to an external buffer (such as an array.array), attempting
          to resize it raises ValueError.
        """
        return len(self._data)

    @capacity.setter
    def capacity(self, value: int) -> None:
        value = arg_to_uint("capacity", value)
        min_capacity = self._start_index + self._sample_count
        if value < min_capacity:
            raise create_capacity_too_small_error(value, min_capacity, "waveform")
        if value != len(self._data):
            if self._data_1d is not None:
                # If _data is a 2D view of a 1D array, resize the base array and recreate the view.
                self._data_1d.resize(value, refcheck=False)
                self._data = self._data_1d.reshape(len(self._data_1d), 1)
            else:
                self._data.resize((value, self.signal_count), refcheck=False)

    @property
    def dtype(self) -> np.dtype[TDigitalState]:
        """The NumPy dtype for the waveform data."""
        return self._data.dtype

    @property
    def extended_properties(self) -> ExtendedPropertyDictionary:
        """The extended properties for the waveform."""
        return self._extended_properties

    @property
    def channel_name(self) -> str:
        """The name of the device channel from which the waveform was acquired."""
        value = self._extended_properties.get(CHANNEL_NAME, "")
        assert isinstance(value, str)
        return value

    @channel_name.setter
    def channel_name(self, value: str) -> None:
        if not isinstance(value, str):
            raise invalid_arg_type("channel name", "str", value)
        self._extended_properties[CHANNEL_NAME] = value

    def _get_signal_names(self) -> list[str]:
        # Lazily allocate self._signal_names if the application needs it.
        signal_names = self._signal_names
        if signal_names is None:
            signal_names_str = self._extended_properties.get(LINE_NAMES, "")
            assert isinstance(signal_names_str, str)
            signal_names = self._signal_names = [
                name.strip() for name in signal_names_str.split(",")
            ]
            if len(signal_names) < self.signal_count:
                signal_names.extend([""] * (self.signal_count - len(signal_names)))
        return signal_names

    def _get_signal_name(self, signal_index: int) -> str:
        return self._get_signal_names()[signal_index]

    def _set_signal_name(self, signal_index: int, value: str) -> None:
        signal_names = self._get_signal_names()
        signal_names[signal_index] = value
        self._extended_properties[LINE_NAMES] = ", ".join(signal_names)

    def _set_timing(self, value: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta]) -> None:
        if self._timing is not value:
            self._timing = value

    def _validate_timing(self, value: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta]) -> None:
        if value._timestamps is not None and len(value._timestamps) != self._sample_count:
            raise create_irregular_timestamp_count_mismatch_error(
                len(value._timestamps), "number of samples in the waveform", self._sample_count
            )

    @property
    def timing(self) -> Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta]:
        """The timing information of the waveform.

        The default value is Timing.empty.
        """
        return self._timing

    @timing.setter
    def timing(self, value: Timing[AnyDateTime, AnyTimeDelta, AnyTimeDelta]) -> None:
        if not isinstance(value, Timing):
            raise invalid_arg_type("timing information", "Timing object", value)
        self._validate_timing(value)
        self._set_timing(value)

    def append(
        self,
        other: (
            npt.NDArray[TDigitalState]
            | DigitalWaveform[TDigitalState]
            | Sequence[DigitalWaveform[TDigitalState]]
        ),
        /,
        timestamps: Sequence[dt.datetime] | Sequence[ht.datetime] | None = None,
    ) -> None:
        """Append data to the waveform.

        Args:
            other: The array or waveform(s) to append.
            timestamps: A sequence of timestamps. When the current waveform has
                SampleIntervalMode.IRREGULAR, you must provide a sequence of timestamps with the
                same length as the array.

        Raises:
            TimingMismatchError: The current and other waveforms have incompatible timing.
            TimingMismatchWarning: The sample intervals of the waveform(s) do not match.
            ValueError: The other array has the wrong number of dimensions or the length of the
                timestamps argument does not match the length of the other array.
            TypeError: The data types of the current waveform and other array or waveform(s) do not
                match, or an argument has the wrong data type.

        When appending waveforms:

        * Timing information is merged based on the sample interval mode of the current
          waveform:

          * SampleIntervalMode.NONE or SampleIntervalMode.REGULAR: The other waveform(s) must also
            have SampleIntervalMode.NONE or SampleIntervalMode.REGULAR. If the sample interval does
            not match, a TimingMismatchWarning is generated. Otherwise, the timing information of
            the other waveform(s) is discarded.

          * SampleIntervalMode.IRREGULAR: The other waveforms(s) must also have
            SampleIntervalMode.IRREGULAR. The timestamps of the other waveforms(s) are appended to
            the current waveform's timing information.

        * Extended properties of the other waveform(s) are merged into the current waveform if they
          are not already set in the current waveform.
        """
        if isinstance(other, np.ndarray):
            self._append_array(other, timestamps)
        elif isinstance(other, DigitalWaveform):
            validate_unsupported_arg("timestamps", timestamps)
            self._append_waveform(other)
        elif isinstance(other, Sequence) and all(isinstance(x, DigitalWaveform) for x in other):
            validate_unsupported_arg("timestamps", timestamps)
            self._append_waveforms(other)
        else:
            raise invalid_arg_type("input", "array or waveform(s)", other)

    def _append_array(
        self,
        array: npt.NDArray[TDigitalState],
        timestamps: Sequence[dt.datetime] | Sequence[ht.datetime] | None = None,
    ) -> None:
        if array.dtype != self.dtype:
            raise create_datatype_mismatch_error("input array", array.dtype, "waveform", self.dtype)

        if array.ndim == 1:
            array_signal_count = 1
            array = array.reshape(len(array), 1)
        elif array.ndim == 2:
            array_signal_count = array.shape[1]
        else:
            raise invalid_array_ndim("input array", "one or two-dimensional array", array.ndim)

        if array_signal_count != self.signal_count:
            raise create_signal_count_mismatch_error(
                "input array", array_signal_count, "waveform", self.signal_count
            )

        if timestamps is not None and len(array) != len(timestamps):
            raise create_irregular_timestamp_count_mismatch_error(
                len(timestamps), "input array length", len(array)
            )

        new_timing = self._timing._append_timestamps(timestamps)

        self._increase_capacity(len(array))
        self._set_timing(new_timing)

        offset = self._start_index + self._sample_count
        self._data[offset : offset + len(array)] = array
        self._sample_count += len(array)

    def _append_waveform(self, waveform: DigitalWaveform[TDigitalState]) -> None:
        self._append_waveforms([waveform])

    def _append_waveforms(self, waveforms: Sequence[DigitalWaveform[TDigitalState]]) -> None:
        for waveform in waveforms:
            if waveform.dtype != self.dtype:
                raise create_datatype_mismatch_error(
                    "input waveform", waveform.dtype, "waveform", self.dtype
                )

        new_timing = self._timing
        for waveform in waveforms:
            new_timing = new_timing._append_timing(waveform._timing)

        self._increase_capacity(sum(waveform.sample_count for waveform in waveforms))
        self._set_timing(new_timing)

        offset = self._start_index + self._sample_count
        for waveform in waveforms:
            self._data[offset : offset + waveform.sample_count] = waveform.data
            offset += waveform.sample_count
            self._sample_count += waveform.sample_count
            self._extended_properties._merge(waveform._extended_properties)

    def _increase_capacity(self, amount: int) -> None:
        new_capacity = self._start_index + self._sample_count + amount
        if new_capacity > self.capacity:
            self.capacity = new_capacity

    def load_data(
        self,
        array: npt.NDArray[TDigitalState],
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
    ) -> None:
        """Load new data into an existing waveform.

        Args:
            array: A NumPy array containing the data to load.
            copy: Specifies whether to copy the array or save a reference to it.
            start_index: The sample index at which the waveform data begins.
            sample_count: The number of samples in the waveform.
        """
        if isinstance(array, np.ndarray):
            self._load_array(array, copy=copy, start_index=start_index, sample_count=sample_count)
        else:
            raise invalid_arg_type("input array", "array", array)

    def _load_array(
        self,
        array: npt.NDArray[TDigitalState],
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        signal_count: SupportsIndex | None = None,
    ) -> None:
        if array.dtype != self.dtype:
            raise create_datatype_mismatch_error("input array", array.dtype, "waveform", self.dtype)

        if array.ndim == 1:
            array_signal_count = 1
            array = array.reshape(len(array), 1)
        elif array.ndim == 2:
            array_signal_count = array.shape[1]
        else:
            raise invalid_array_ndim("input array", "one or two-dimensional array", array.ndim)

        if self._timing._timestamps is not None and len(array) != len(self._timing._timestamps):
            raise create_irregular_timestamp_count_mismatch_error(
                len(self._timing._timestamps), "input array length", len(array), reversed=True
            )

        start_index = arg_to_uint("start index", start_index, 0)
        sample_count = arg_to_uint("sample count", sample_count, len(array) - start_index)
        signal_count = arg_to_uint("signal count", signal_count, array_signal_count)

        if signal_count != array_signal_count:
            raise create_signal_count_mismatch_error(
                "input array", signal_count, "waveform", array_signal_count
            )

        if copy:
            if sample_count > len(self._data):
                self.capacity = sample_count
            self._data[0:sample_count] = array[start_index : start_index + sample_count]
            self._start_index = 0
            self._sample_count = sample_count
        else:
            self._data = array
            self._start_index = start_index
            self._sample_count = sample_count

    def test(
        self,
        expected_waveform: DigitalWaveform[TDigitalState],
        *,
        start_sample: SupportsIndex | None = 0,
        expected_start_sample: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
    ) -> DigitalWaveformTestResult:
        """Test the digital waveform against an expected digital waveform.

        Args:
            expected_waveform: The expected digital waveform to compare against.
            start_sample: The beginning sample of ``self`` to compare.
            expected_start_sample: The beginning sample of ``expected_waveform`` to compare.
            sample_count: The number of samples to compare.

        Returns:
            The test result.
        """
        start_sample = arg_to_uint("start sample", start_sample, 0)
        expected_start_sample = arg_to_uint("expected start sample", expected_start_sample, 0)
        sample_count = arg_to_uint("sample count", sample_count, self.sample_count - start_sample)

        if self.signal_count != expected_waveform.signal_count:
            raise create_signal_count_mismatch_error(
                "expected waveform", expected_waveform.signal_count, "waveform", self.signal_count
            )
        if start_sample + sample_count > self.sample_count:
            raise create_start_index_or_sample_count_too_large_error(
                start_sample, sample_count, "number of samples in the waveform", self.sample_count
            )
        if expected_start_sample + sample_count > expected_waveform.sample_count:
            raise create_start_index_or_sample_count_too_large_error(
                expected_start_sample,
                sample_count,
                "number of samples in the expected waveform",
                expected_waveform.sample_count,
            )

        failures = []
        for _ in range(sample_count):
            for signal_index in range(self.signal_count):
                actual_state = DigitalState(self.data[start_sample, signal_index])
                expected_state = DigitalState(
                    expected_waveform.data[expected_start_sample, signal_index]
                )
                if DigitalState.test(actual_state, expected_state):
                    failures.append(
                        DigitalWaveformFailure(
                            start_sample,
                            expected_start_sample,
                            signal_index,
                            actual_state,
                            expected_state,
                        )
                    )
            start_sample += 1
            expected_start_sample += 1

        return DigitalWaveformTestResult(failures)

    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        return (
            self.dtype == value.dtype
            and np.array_equal(self.data, value.data)
            and self._extended_properties == value._extended_properties
            and self._timing == value._timing
        )

    def __reduce__(self) -> tuple[Any, ...]:
        """Return object state for pickling."""
        ctor_args = (self._sample_count, self.signal_count, self.dtype)
        ctor_kwargs: dict[str, Any] = {
            "data": self.data,
            "extended_properties": self._extended_properties,
            "copy_extended_properties": False,
            "timing": self._timing,
        }
        return (self.__class__._unpickle, (ctor_args, ctor_kwargs))

    @classmethod
    def _unpickle(cls, args: tuple[Any, ...], kwargs: dict[str, Any]) -> Self:
        return cls(*args, **kwargs)

    def __repr__(self) -> str:
        """Return repr(self)."""
        args = [f"{self._sample_count}, {self.signal_count}"]
        if self.dtype != np.uint8:
            args.append(f"{self.dtype.name}")
        # start_index and capacity are not shown because they are allocation details. data hides
        # the unused data before start_index and after start_index+sample_count.
        if self._sample_count > 0:
            # Hack: undo NumPy's line wrapping
            args.append(f"data={self.data!r}".replace("\n      ", ""))
        if self._extended_properties:
            args.append(f"extended_properties={self._extended_properties._properties!r}")
        if self._timing is not Timing.empty:
            args.append(f"timing={self._timing!r}")
        return f"{self.__class__.__module__}.{self.__class__.__name__}({', '.join(args)})"



from __future__ import annotations

import datetime as dt
import decimal
import math
import operator
from decimal import Decimal
from functools import singledispatchmethod
from typing import TYPE_CHECKING, Any, ClassVar, SupportsIndex, Union, final, overload

import hightime as ht
from typing_extensions import Self, TypeAlias

from nitypes._arguments import arg_to_int
from nitypes._exceptions import int_out_of_range, invalid_arg_type

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.bintime import TimeValueTuple
else:
    from nitypes.bintime._time_value_tuple import TimeValueTuple


_INT64_MAX = (1 << 63) - 1
_INT64_MIN = -(1 << 63)
_UINT64_MAX = (1 << 64) - 1
_UINT64_MIN = 0
_INT128_MAX = (1 << 127) - 1
_INT128_MIN = -(1 << 127)

_BITS_PER_SECOND = 64
_TICKS_PER_SECOND = 1 << _BITS_PER_SECOND
_FRACTIONAL_SECONDS_MASK = _TICKS_PER_SECOND - 1

_SECONDS_PER_DAY = 86400

_MICROSECONDS_PER_SECOND = 10**6
_FEMTOSECONDS_PER_SECOND = 10**15
_YOCTOSECONDS_PER_SECOND = 10**24

_FEMTOSECONDS_PER_MICROSECOND = _FEMTOSECONDS_PER_SECOND // _MICROSECONDS_PER_SECOND
_YOCTOSECONDS_PER_FEMTOSECOND = _YOCTOSECONDS_PER_SECOND // _FEMTOSECONDS_PER_SECOND

_DECIMAL_DIGITS = 64
_REPR_TICKS = False


_OtherTimeDelta: TypeAlias = Union[dt.timedelta, ht.timedelta]
_OTHER_TIMEDELTA_TUPLE = (dt.timedelta, ht.timedelta)

_OtherDateTime: TypeAlias = Union[dt.datetime, ht.datetime]


@final
class TimeDelta:
    """A duration, represented in NI Binary Time Format (NI-BTF).

    TimeDelta represents time as a 128-bit fixed point number with 64-bit whole seconds and 64-bit
    fractional seconds.

    .. warning::
        The fractional seconds are represented as a binary fraction, which is a sum of inverse
        powers of 2. Values that are not exactly representable as binary fractions will display
        rounding error or "bruising" similar to a floating point number.

    TimeDelta instances are duck typing compatible with :any:`datetime.timedelta` and
    :any:`hightime.timedelta`.

    Constructing
    ^^^^^^^^^^^^

    You can construct a :class:`TimeDelta` from a number of seconds, expressed as an :any:`int`,
    :any:`float`, or :any:`decimal.Decimal`.

    >>> TimeDelta(100)
    nitypes.bintime.TimeDelta(Decimal('100'))
    >>> TimeDelta(100.125)
    nitypes.bintime.TimeDelta(Decimal('100.125'))
    >>> from decimal import Decimal
    >>> TimeDelta(Decimal("100.125"))
    nitypes.bintime.TimeDelta(Decimal('100.125'))

    :class:`TimeDelta` has the same resolution and rounding behavior as :class:`DateTime`.

    >>> TimeDelta(Decimal("100.01234567890123456789"))
    nitypes.bintime.TimeDelta(Decimal('100.012345678901234567889'))

    Unlike other ``timedelta`` objects, you cannot construct a :class:`TimeDelta` from separate weeks,
    days, hours, etc. If you want to do that, construct a :any:`datetime.timedelta` or
    :any:`hightime.timedelta` and then use it to construct a :class:`TimeDelta`.

    >>> import datetime, hightime
    >>> TimeDelta(datetime.timedelta(days=1, microseconds=1))
    nitypes.bintime.TimeDelta(Decimal('86400.0000010000000000000'))
    >>> TimeDelta(hightime.timedelta(days=1, femtoseconds=1))
    nitypes.bintime.TimeDelta(Decimal('86400.0000000000000010000'))

    Math Operations
    ^^^^^^^^^^^^^^^

    :class:`DateTime` and :class:`TimeDelta` support the same math operations as :any:`datetime.datetime`
    and :any:`datetime.timedelta`.

    For example, you can add or subtract :class:`TimeDelta` objects together:

    >>> TimeDelta(100.5) + TimeDelta(0.5)
    nitypes.bintime.TimeDelta(Decimal('101'))
    >>> TimeDelta(100.5) - TimeDelta(0.5)
    nitypes.bintime.TimeDelta(Decimal('100'))

    Or add/subtract a :class:`DateTime` with a :class:`TimeDelta`, :any:`datetime.timedelta`, or
    :any:`hightime.timedelta`:

    >>> DateTime(2025, 1, 1, tzinfo=datetime.timezone.utc) + TimeDelta(86400)
    nitypes.bintime.DateTime(2025, 1, 2, 0, 0, tzinfo=datetime.timezone.utc)
    >>> DateTime(2025, 1, 1, tzinfo=datetime.timezone.utc) + datetime.timedelta(days=1)
    nitypes.bintime.DateTime(2025, 1, 2, 0, 0, tzinfo=datetime.timezone.utc)
    >>> DateTime(2025, 1, 1, tzinfo=datetime.timezone.utc) + hightime.timedelta(femtoseconds=1)
    nitypes.bintime.DateTime(2025, 1, 1, 0, 0, 0, 0, 1, 13873, tzinfo=datetime.timezone.utc)

    Class members
    ^^^^^^^^^^^^^
    """  # noqa: W505 - doc line too long

    min: ClassVar[TimeDelta]
    """The most negative :class:`TimeDelta` object, approximately -292 million years."""

    max: ClassVar[TimeDelta]
    """The most positive :class:`TimeDelta` object, approximately 292 million years."""

    __slots__ = ["_ticks"]

    _ticks: int

    @overload
    def __init__(self) -> None: ...  # noqa: D107 - missing docstring in __init__

    @overload
    def __init__(  # noqa: D107 - missing docstring in __init__
        self, value: _OtherTimeDelta, /
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - missing docstring in __init__
        self, seconds: SupportsIndex | Decimal | float
    ) -> None: ...

    def __init__(
        self,
        seconds: SupportsIndex | Decimal | float | _OtherTimeDelta | None = None,
    ) -> None:
        """Initialize a TimeDelta."""
        ticks = self.__class__._to_ticks(seconds)
        if not (_INT128_MIN <= ticks <= _INT128_MAX):
            raise OverflowError(
                "The seconds value is out of range.\n\n"
                f"Requested value: {seconds}\n"
                f"Minimum value: {self.__class__.min.precision_total_seconds()}\n"
                f"Maximum value: {self.__class__.max.precision_total_seconds()}"
            )
        self._ticks = ticks

    @singledispatchmethod
    @classmethod
    def _to_ticks(cls, seconds: object) -> int:
        raise invalid_arg_type("seconds", "number or timedelta", seconds)

    # Python 3.9: pass the type to register() in order to work around
    # https://github.com/python/cpython/issues/86153 - singledispatchmethod raises an error when
    # relying on a forward declaration
    @_to_ticks.register(SupportsIndex)
    @classmethod
    def _(cls, seconds: SupportsIndex) -> int:
        return operator.index(seconds) << _BITS_PER_SECOND

    @_to_ticks.register(Decimal)
    @classmethod
    def _(cls, seconds: Decimal) -> int:
        with decimal.localcontext() as ctx:
            ctx.prec = _DECIMAL_DIGITS
            whole_seconds, fractional_seconds = divmod(seconds, 1)
            ticks = int(whole_seconds) * _TICKS_PER_SECOND
            ticks += round(fractional_seconds * _TICKS_PER_SECOND)
            return ticks

    @_to_ticks.register(float)
    @classmethod
    def _(cls, seconds: float) -> int:
        fractional_seconds, whole_seconds = math.modf(seconds)
        ticks = int(whole_seconds) * _TICKS_PER_SECOND
        ticks += round(fractional_seconds * _TICKS_PER_SECOND)
        return ticks

    @_to_ticks.register(ht.timedelta)
    @classmethod
    def _(cls, seconds: ht.timedelta) -> int:
        return cls._to_ticks(seconds.precision_total_seconds())

    @_to_ticks.register(dt.timedelta)
    @classmethod
    def _(cls, seconds: dt.timedelta) -> int:
        # Do not use total_seconds() because it loses precision.
        ticks = (seconds.days * _SECONDS_PER_DAY) << _BITS_PER_SECOND
        ticks += seconds.seconds << _BITS_PER_SECOND
        ticks += (seconds.microseconds << _BITS_PER_SECOND) // _MICROSECONDS_PER_SECOND
        return ticks

    @_to_ticks.register(type(None))
    @classmethod
    def _(cls, seconds: None) -> int:
        return 0

    @classmethod
    def from_ticks(cls, ticks: SupportsIndex) -> Self:
        """Create a TimeDelta from a 128-bit fixed point number expressed as an integer."""
        ticks = arg_to_int("ticks", ticks)
        if not (_INT128_MIN <= ticks <= _INT128_MAX):
            raise int_out_of_range(ticks, _INT128_MIN, _INT128_MAX)
        self = cls.__new__(cls)
        self._ticks = ticks
        return self

    @classmethod
    def from_tuple(cls, value: TimeValueTuple) -> Self:
        """Create a TimeDelta from 64-bit whole seconds and fractional seconds ints."""
        if not (_INT64_MIN <= value.whole_seconds <= _INT64_MAX):
            raise int_out_of_range(value.whole_seconds, _INT64_MIN, _INT64_MAX)
        if not (_UINT64_MIN <= value.fractional_seconds <= _UINT64_MAX):
            raise int_out_of_range(value.fractional_seconds, _UINT64_MIN, _UINT64_MAX)
        ticks = value.whole_seconds << _BITS_PER_SECOND
        ticks = ticks | value.fractional_seconds
        return cls.from_ticks(ticks)

    def _to_datetime_timedelta(self) -> dt.timedelta:
        """Return self as a :any:`datetime.timedelta`."""
        whole_seconds = self._ticks >> _BITS_PER_SECOND
        microseconds = (
            _MICROSECONDS_PER_SECOND * (self._ticks & _FRACTIONAL_SECONDS_MASK)
        ) >> _BITS_PER_SECOND
        return dt.timedelta(seconds=whole_seconds, microseconds=microseconds)

    def _to_hightime_timedelta(self) -> ht.timedelta:
        """Return self as a :any:`hightime.timedelta`."""
        whole_seconds = self._ticks >> _BITS_PER_SECOND
        yoctoseconds = (
            _YOCTOSECONDS_PER_SECOND * (self._ticks & _FRACTIONAL_SECONDS_MASK)
        ) >> _BITS_PER_SECOND
        return ht.timedelta(seconds=whole_seconds, yoctoseconds=yoctoseconds)

    @property
    def days(self) -> int:
        """The number of days in the time delta."""
        return (self._ticks >> _BITS_PER_SECOND) // _SECONDS_PER_DAY

    @property
    def seconds(self) -> int:
        """The number of seconds in the time delta, up to the nearest day."""
        return (self._ticks >> _BITS_PER_SECOND) % _SECONDS_PER_DAY

    @property
    def microseconds(self) -> int:
        """The number of microseconds in the time delta, up to the nearest second."""
        return (
            _MICROSECONDS_PER_SECOND * (self._ticks & _FRACTIONAL_SECONDS_MASK)
        ) >> _BITS_PER_SECOND

    @property
    def femtoseconds(self) -> int:
        """The number of femtoseconds in the time delta, up to the nearest microsecond."""
        value = (
            _FEMTOSECONDS_PER_SECOND * (self._ticks & _FRACTIONAL_SECONDS_MASK) >> _BITS_PER_SECOND
        )
        return value % _FEMTOSECONDS_PER_MICROSECOND

    @property
    def yoctoseconds(self) -> int:
        """The number of yoctoseconds in the time delta, up to the nearest femtosecond.

        .. warning::
            Because this class uses a 64-bit binary fraction, the smallest value it can represent
            is ``1.0 / (1 << 64)`` seconds, which is about 54210 yoctoseconds.
        """
        value = (
            _YOCTOSECONDS_PER_SECOND * (self._ticks & _FRACTIONAL_SECONDS_MASK)
        ) >> _BITS_PER_SECOND
        return value % _YOCTOSECONDS_PER_FEMTOSECOND

    @property
    def ticks(self) -> int:
        """The total ticks in the time delta as a 128-bit integer."""
        return self._ticks

    def to_tuple(self) -> TimeValueTuple:
        """The whole seconds and fractional seconds parts of the time delta as 64-bit ints."""
        whole_seconds = self._ticks >> _BITS_PER_SECOND
        fractional_seconds = self._ticks & _FRACTIONAL_SECONDS_MASK
        return TimeValueTuple(whole_seconds, fractional_seconds)

    def total_seconds(self) -> float:
        """The total seconds in the time delta.

        .. warning::
            Converting a time value to a floating point number loses precision. Consider using
            :any:`precision_total_seconds` instead.
        """
        seconds = float(self._ticks >> _BITS_PER_SECOND)
        seconds += float((self._ticks & _FRACTIONAL_SECONDS_MASK) / _TICKS_PER_SECOND)
        return seconds

    def precision_total_seconds(self) -> Decimal:
        """The precise total seconds in the time delta.

        Note: up to 64 significant digits are used in computation.
        """
        with decimal.localcontext() as ctx:
            ctx.prec = _DECIMAL_DIGITS
            seconds = Decimal(self._ticks >> _BITS_PER_SECOND)
            seconds += Decimal(self._ticks & _FRACTIONAL_SECONDS_MASK) / Decimal(_TICKS_PER_SECOND)
            return seconds

    def __neg__(self) -> TimeDelta:
        """Return -self."""
        return self.__class__.from_ticks(-self._ticks)

    def __pos__(self) -> TimeDelta:
        """Return +self."""
        return self

    def __abs__(self) -> TimeDelta:
        """Return abs(self)."""
        return -self if self._ticks < 0 else self

    @overload
    def __add__(  # noqa: D105 - missing docstring in magic method
        self, value: TimeDelta | _OtherTimeDelta, /
    ) -> TimeDelta: ...

    @overload
    def __add__(  # noqa: D105 - missing docstring in magic method
        self, value: ht.datetime, /
    ) -> ht.datetime: ...

    @overload
    def __add__(  # noqa: D105 - missing docstring in magic method
        self, value: dt.datetime, /
    ) -> dt.datetime: ...

    def __add__(
        self, value: TimeDelta | _OtherTimeDelta | _OtherDateTime, /
    ) -> TimeDelta | _OtherDateTime:
        """Return self+value."""
        if isinstance(value, TimeDelta):
            return self.__class__.from_ticks(self._ticks + value._ticks)
        elif isinstance(value, _OTHER_TIMEDELTA_TUPLE):
            return self + self.__class__(value)
        elif isinstance(value, ht.datetime):
            return self._to_hightime_timedelta() + value
        # Handle dt.datetime separately to round to the nearest microsecond instead of truncating.
        elif isinstance(value, dt.datetime):
            return self._to_datetime_timedelta() + value
        else:
            return NotImplemented

    __radd__ = __add__

    # __sub__ doesn't support _TOtherDateTime because it doesn't make sense to negate a datetime.
    def __sub__(self, value: TimeDelta | _OtherTimeDelta, /) -> TimeDelta:
        """Return self-value."""
        if isinstance(value, TimeDelta):
            return self.__class__.from_ticks(self._ticks - value._ticks)
        elif isinstance(value, _OTHER_TIMEDELTA_TUPLE):
            return self - self.__class__(value)
        else:
            return NotImplemented

    # __rsub__ supports _TOtherDateTime in order to support subtracting a bintime.TimeDelta from a
    # datetime.datetime or hightime.datetime.
    @overload
    def __rsub__(  # noqa: D105 - missing docstring in magic method
        self, value: TimeDelta | _OtherTimeDelta, /
    ) -> TimeDelta: ...

    @overload
    def __rsub__(  # noqa: D105 - missing docstring in magic method
        self, value: ht.datetime, /
    ) -> ht.datetime: ...

    @overload
    def __rsub__(  # noqa: D105 - missing docstring in magic method
        self, value: dt.datetime, /
    ) -> dt.datetime: ...

    # AB#3153350 - nitypes.bintime.DateTime signatures of __sub__ and __rsub__ are unsafely
    # overlapping
    def __rsub__(  # type: ignore[misc]
        self, value: TimeDelta | _OtherTimeDelta | _OtherDateTime, /
    ) -> TimeDelta | _OtherDateTime:
        """Return value-self."""
        if isinstance(value, TimeDelta):
            return self.__class__.from_ticks(value._ticks - self._ticks)
        elif isinstance(value, _OTHER_TIMEDELTA_TUPLE):
            return self.__class__(value) - self
        elif isinstance(value, ht.datetime):
            return value - self._to_hightime_timedelta()
        # Handle dt.datetime separately to round to the nearest microsecond instead of truncating.
        elif isinstance(value, dt.datetime):
            return value - self._to_datetime_timedelta()
        else:
            return NotImplemented

    def __mul__(self, value: int | float | Decimal, /) -> TimeDelta:
        """Return self*value."""
        if isinstance(value, int):
            return self.__class__.from_ticks(self._ticks * value)
        elif isinstance(value, float):
            # Using floating point math on 128-bit numbers loses precision, so use Decimal math.
            return self * Decimal(value)
        elif isinstance(value, Decimal):
            with decimal.localcontext() as ctx:
                ctx.prec = _DECIMAL_DIGITS
                return self.__class__(self.precision_total_seconds() * value)
        else:
            return NotImplemented

    __rmul__ = __mul__

    @overload
    def __floordiv__(  # noqa: D105 - missing docstring in magic method
        self, value: TimeDelta, /
    ) -> int: ...
    @overload
    def __floordiv__(  # noqa: D105 - missing docstring in magic method
        self, value: int, /
    ) -> TimeDelta: ...

    def __floordiv__(self, value: TimeDelta | int, /) -> int | TimeDelta:
        """Return self//value."""
        if isinstance(value, TimeDelta):
            return self._ticks // value._ticks
        elif isinstance(value, int):
            return self.__class__.from_ticks(self._ticks // value)
        else:
            return NotImplemented

    @overload
    def __truediv__(  # noqa: D105 - missing docstring in magic method
        self, value: TimeDelta, /
    ) -> float: ...
    @overload
    def __truediv__(  # noqa: D105 - missing docstring in magic method
        self, value: float, /
    ) -> TimeDelta: ...

    def __truediv__(self, value: TimeDelta | float, /) -> float | TimeDelta:
        """Return self/value."""
        if isinstance(value, TimeDelta):
            return self.total_seconds() / value.total_seconds()
        elif isinstance(value, float):
            return self.__class__(self.total_seconds() / value)
        else:
            return NotImplemented

    def __mod__(self, value: TimeDelta | _OtherTimeDelta, /) -> TimeDelta:
        """Return self%value."""
        if isinstance(value, TimeDelta):
            return self.__class__.from_ticks(self._ticks % value._ticks)
        elif isinstance(value, _OTHER_TIMEDELTA_TUPLE):
            return self % self.__class__(value)
        else:
            return NotImplemented

    def __divmod__(self, value: TimeDelta | _OtherTimeDelta, /) -> tuple[int, TimeDelta]:
        """Return (self//value, self%value)."""
        if isinstance(value, TimeDelta):
            return (self // value, self % value)
        elif isinstance(value, _OTHER_TIMEDELTA_TUPLE):
            return divmod(self, self.__class__(value))
        else:
            return NotImplemented

    # In comparison operators, always promote to the more precise data type (dt -> bt, bt -> ht).
    def __lt__(self, value: TimeDelta | _OtherTimeDelta, /) -> bool:
        """Return self<value."""
        if isinstance(value, self.__class__):
            return self._ticks < value._ticks
        elif isinstance(value, ht.timedelta):
            return self._to_hightime_timedelta() < value
        elif isinstance(value, dt.timedelta):
            return self < self.__class__(value)
        else:
            return NotImplemented

    def __le__(self, value: TimeDelta | _OtherTimeDelta, /) -> bool:
        """Return self<=value."""
        if isinstance(value, self.__class__):
            return self._ticks <= value._ticks
        elif isinstance(value, ht.timedelta):
            return self._to_hightime_timedelta() <= value
        elif isinstance(value, dt.timedelta):
            return self <= self.__class__(value)
        else:
            return NotImplemented

    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        if isinstance(value, self.__class__):
            return self._ticks == value._ticks
        elif isinstance(value, ht.timedelta):
            return self._to_hightime_timedelta() == value
        elif isinstance(value, dt.timedelta):
            return self == self.__class__(value)
        else:
            return NotImplemented

    def __gt__(self, value: TimeDelta | _OtherTimeDelta, /) -> bool:
        """Return self<value."""
        if isinstance(value, self.__class__):
            return self._ticks > value._ticks
        elif isinstance(value, ht.timedelta):
            return self._to_hightime_timedelta() > value
        elif isinstance(value, dt.timedelta):
            return self > self.__class__(value)
        else:
            return NotImplemented

    def __ge__(self, value: TimeDelta | _OtherTimeDelta, /) -> bool:
        """Return self>=value."""
        if isinstance(value, self.__class__):
            return self._ticks >= value._ticks
        elif isinstance(value, ht.timedelta):
            return self._to_hightime_timedelta() >= value
        elif isinstance(value, dt.timedelta):
            return self >= self.__class__(value)
        else:
            return NotImplemented

    def __bool__(self) -> bool:
        """Return bool(self)."""
        return self._ticks != 0

    def __hash__(self) -> int:
        """Return hash(self)."""
        return hash(self._ticks)

    def __reduce__(self) -> tuple[Any, ...]:
        """Return object state for pickling."""
        return (self.__class__.from_ticks, (self._ticks,))

    def __str__(self) -> str:
        """Return repr(self)."""
        days = self.days
        seconds = self.seconds
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)
        # Display up to 18 digits of fractional seconds, rounded to the nearest digit.
        fractional_seconds = 10**18 * (self._ticks & _FRACTIONAL_SECONDS_MASK)
        fractional_seconds = (fractional_seconds + _TICKS_PER_SECOND // 2) // _TICKS_PER_SECOND
        s = f"{days} day, " if abs(days) == 1 else f"{days} days, " if days else ""
        s += f"{hours}:{minutes:02}:{seconds:02}"
        if fractional_seconds != 0:
            s += f".{fractional_seconds:018}".rstrip("0")  # strip trailing zeroes
        return s

    def __repr__(self) -> str:
        """Return repr(self)."""
        if _REPR_TICKS:
            return (
                f"{self.__class__.__module__}.{self.__class__.__name__}"
                f".from_ticks({self._ticks})"
            )
        # Display up to 24 decimal digits (yoctoseconds), like hightime does. The smallest time
        # increment representable with NI-BTF is 54210 yoctoseconds, so if all 24 decimal digits are
        # displayed, the last few are due to rounding error.
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}"
            f"(Decimal('{self.precision_total_seconds():.24}'))"
        )


TimeDelta.max = TimeDelta.from_ticks(_INT128_MAX)
TimeDelta.min = TimeDelta.from_ticks(_INT128_MIN)



r"""Binary time data types for NI Python APIs.

NI Binary Time Format
=====================

This module implements the NI Binary Time Format (`NI-BTF`_), a high-resolution time format used by
NI software.

An NI-BTF time value is a 128-bit fixed point number consisting of a 64-bit whole seconds part and a
64-bit fractional seconds part. There are two types of NI-BTF time values:

* An NI-BTF absolute time represents a point in time as the number of seconds after midnight,
  January 1, 1904, UTC.
* An NI-BTF time interval represents a difference between two points in time.

NI-BTF time types are also supported in LabVIEW, LabWindows/CVI, and .NET. You can use NI-BTF time
to efficiently share high-resolution date-time information with other NI application development
environments.

.. _ni-btf: https://www.ni.com/docs/en-US/bundle/labwindows-cvi/page/cvi/libref/ni-btf.htm

NI-BTF Python Data Types
========================

* :class:`DateTime`: represents an NI-BTF absolute time as a Python object.
* :class:`TimeDelta`: represents a NI-BTF time interval as a Python object.
* :class:`TimeDeltaArray`: an array of :class:`TimeDelta` values.

NI-BTF NumPy Structured Data Types
==================================

:any:`CVIAbsoluteTimeDType` and :any:`CVITimeIntervalDType` are NumPy structured data type objects
representing the ``CVIAbsoluteTime`` and ``CVITimeInterval`` C structs. These structured data types
can be used to efficiently represent NI-BTF time values in NumPy arrays or pass them to/from C DLLs.

.. warning::
    :any:`CVIAbsoluteTimeDType` and :any:`CVITimeIntervalDType` have the same layout and field
    names, so NumPy and type checkers such as Mypy currently treat them as the same type.

NI-BTF versus ``hightime``
==========================

NI also provides the ``hightime`` Python package, which extends the standard Python :mod:`datetime`
module to support up to yoctosecond precision.

``nitypes.bintime`` is not a replacement for ``hightime``. The two time formats have different
strengths and weaknesses.

* ``hightime`` supports local time zones and time-zone-naive times. ``bintime`` only supports UTC.
* ``hightime`` classes supports the same operations as the standard ``datetime`` classes.
  ``bintime`` classes support a subset of the standard ``datetime`` operations.
* ``hightime`` has a larger memory footprint than NI-BTF. ``hightime`` objects are separately
  allocated from the heap. ``bintime`` offers the choice of separately allocated Python objects or
  a more compact NumPy representation that can store multiple timestamps in a single block of
  memory.
* ``hightime`` requires conversion to/from NI-BTF when calling the NI driver C APIs from Python.
  ``nitypes.bintime`` includes reusable conversion routines for NI driver Python APIs to use.

NI-BTF versus :any:`numpy.datetime64`
=====================================

NumPy provides the :any:`numpy.datetime64` data type, which is even more compact than NI-BTF.
However, it has lower resolution than NI-BTF and is not interoperable with NI driver C APIs that use
NI-BTF.
"""

from __future__ import annotations

from nitypes.bintime._datetime import DateTime
from nitypes.bintime._dtypes import (
    CVIAbsoluteTimeBase,
    CVIAbsoluteTimeDType,
    CVITimeIntervalBase,
    CVITimeIntervalDType,
)
from nitypes.bintime._time_value_tuple import TimeValueTuple
from nitypes.bintime._timedelta import TimeDelta
from nitypes.bintime._timedelta_array import TimeDeltaArray

__all__ = [
    "DateTime",
    "CVIAbsoluteTimeBase",
    "CVIAbsoluteTimeDType",
    "CVITimeIntervalBase",
    "CVITimeIntervalDType",
    "TimeDelta",
    "TimeDeltaArray",
    "TimeValueTuple",
]

# Hide that it was defined in a helper file
DateTime.__module__ = __name__
TimeDelta.__module__ = __name__
TimeDeltaArray.__module__ = __name__
TimeValueTuple.__module__ = __name__



from __future__ import annotations

import datetime as dt

import hightime as ht

import nitypes.bintime as bt

ANY_DATETIME_TUPLE = (bt.DateTime, dt.datetime, ht.datetime)
"""Tuple of types corresponding to :any:`AnyDateTime`."""

ANY_TIMEDELTA_TUPLE = (bt.TimeDelta, dt.timedelta, ht.timedelta)
"""Tuple of types corresponding to :any:`AnyTimeDelta`."""



from __future__ import annotations

from nitypes.waveform.warnings import TimingMismatchWarning, ScalingMismatchWarning


def sample_interval_mismatch() -> TimingMismatchWarning:
    """Create a TimingMismatchWarning about appending waveforms with mismatched sample intervals."""
    return TimingMismatchWarning(
        "The sample interval of one or more waveforms does not match the sample interval of the current waveform."
    )


def scale_mode_mismatch() -> ScalingMismatchWarning:
    """Create a ScalingMismatchwarning about appending waveforms with mismatched scale modes."""
    return ScalingMismatchWarning(
        "The scale mode of one or more waveforms does not match the scale mode of the current waveform."
    )



"""Vector data type for NI Python APIs.

Vector Data Type
=================
:class:`Vector`: A vector data object represents an array of scalar values with units information.
Valid types for the scalar value are :any:`bool`, :any:`int`, :any:`float`, and :any:`str`.
"""

from __future__ import annotations

from collections.abc import Iterable, MutableSequence
from typing import TYPE_CHECKING, overload, Any, Union

from typing_extensions import TypeVar, final, override

from nitypes._exceptions import invalid_arg_type
from nitypes.waveform._extended_properties import UNIT_DESCRIPTION

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import ExtendedPropertyDictionary
else:
    from nitypes.waveform._extended_properties import ExtendedPropertyDictionary

TScalar = TypeVar("TScalar", bound=Union[bool, int, float, str])


@final
class Vector(MutableSequence[TScalar]):
    """A sequence of scalar values with units information.

    Constructing
    ^^^^^^^^^^^^

    To construct a vector data object, use the :class:`Vector` class:

    >>> Vector([False, True])
    nitypes.vector.Vector(values=[False, True], units='')
    >>> Vector([0, 1, 2])
    nitypes.vector.Vector(values=[0, 1, 2], units='')
    >>> Vector([5.0, 6.0], 'volts')
    nitypes.vector.Vector(values=[5.0, 6.0], units='volts')
    >>> Vector(["one", "two"], "volts")
    nitypes.vector.Vector(values=['one', 'two'], units='volts')
    """

    __slots__ = [
        "_values",
        "_value_type",
        "_extended_properties",
    ]

    _values: list[TScalar]
    _value_type: type[TScalar]
    _extended_properties: ExtendedPropertyDictionary

    def __init__(
        self,
        values: Iterable[TScalar],
        units: str = "",
        *,
        value_type: type[TScalar] | None = None,
    ) -> None:
        """Initialize a new vector.

        Args:
            values: The scalar values to store in this object.
            units: The units string associated with this data.
            value_type: The type of values that will be added to this Vector.
                This parameter should only be used when creating a Vector with
                an empty Iterable.

        Returns:
            A vector data object.
        """
        if not values:
            if not value_type:
                raise TypeError("You must specify values as non-empty or specify value_type.")
            self._value_type = value_type
        else:
            # Validate the values input
            for index, value in enumerate(values):
                # Only set _value_type once.
                if not index:
                    self._value_type = type(value)

                if not isinstance(value, (bool, int, float, str)):
                    raise invalid_arg_type("vector input data", "bool, int, float, or str", values)

                if not isinstance(value, self._value_type):
                    raise TypeError("All values in the values input must be of the same type.")

        if not isinstance(units, str):
            raise invalid_arg_type("units", "str", units)

        self._values = list(values)
        self._extended_properties = ExtendedPropertyDictionary()
        self._extended_properties[UNIT_DESCRIPTION] = units

    @property
    def units(self) -> str:
        """The unit of measurement, such as volts, of the vector."""
        value = self._extended_properties.get(UNIT_DESCRIPTION, "")
        assert isinstance(value, str)
        return value

    @units.setter
    def units(self, value: str) -> None:
        if not isinstance(value, str):
            raise invalid_arg_type("units", "str", value)
        self._extended_properties[UNIT_DESCRIPTION] = value

    @property
    def extended_properties(self) -> ExtendedPropertyDictionary:
        """The extended properties for the vector.

        .. note::
            Data stored in the extended properties dictionary may not be encrypted when you send it
            over the network or write it to a TDMS file.
        """
        return self._extended_properties

    @overload
    def __getitem__(  # noqa: D105 - missing docstring in magic method
        self, index: int
    ) -> TScalar: ...

    @overload
    def __getitem__(  # noqa: D105 - missing docstring in magic method
        self, index: slice
    ) -> MutableSequence[TScalar]: ...

    @override
    def __getitem__(self, index: int | slice) -> TScalar | MutableSequence[TScalar]:
        """Return the TimeDelta at the specified location."""
        return self._values[index]

    @overload
    def __setitem__(  # noqa: D105 - missing docstring in magic method
        self, index: int, value: TScalar
    ) -> None: ...

    @overload
    def __setitem__(  # noqa: D105 - missing docstring in magic method
        self, index: slice, value: Iterable[TScalar]
    ) -> None: ...

    @override
    def __setitem__(self, index: int | slice, value: TScalar | Iterable[TScalar]) -> None:
        """Set value(s) at the specified location."""
        if isinstance(index, int):
            if isinstance(value, Iterable) and not isinstance(value, str):
                raise TypeError("You cannot assign an Iterable to a vector index.")
            elif not isinstance(value, self._value_type):
                raise self._create_value_mismatch_exception(value)

            self._values[index] = value
        else:  # slice
            if not isinstance(value, Iterable):
                raise TypeError("You must assign an Iterable to a Vector slice.")
            elif isinstance(value, str):  # Narrow the type to exclude string.
                raise TypeError("You cannot assign a string to Vector slice.")
            else:
                # Assigning an empty Iterable to a slice is valid, so we don't check for empty.
                # If an empty Iterable is assigned to a slice, that slice is deleted.
                for subval in value:
                    if not isinstance(subval, self._value_type):
                        raise self._create_value_mismatch_exception(subval)

            self._values[index] = value

    def __delitem__(self, index: int | slice) -> None:
        """Delete item(s) from the specified location."""
        del self._values[index]

    def __len__(self) -> int:
        """Return the length of the Vector."""
        return len(self._values)

    def insert(self, index: int, value: TScalar) -> None:
        """Insert a value at the specified location."""
        if not isinstance(value, self._value_type):
            raise self._create_value_mismatch_exception(value)
        self._values.insert(index, value)

    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        return self._values == value._values and self.units == value.units

    def __reduce__(self) -> tuple[Any, ...]:
        """Return object state for pickling."""
        return (self.__class__, (self._values, self.units))

    def __repr__(self) -> str:
        """Return repr(self)."""
        args = [f"values={self._values!r}", f"units={self.units!r}"]
        return f"{self.__class__.__module__}.{self.__class__.__name__}({', '.join(args)})"

    def __str__(self) -> str:
        """Return str(self)."""
        if self.units:
            values_with_units = [f"{value} {self.units}" for value in self._values]
            comma_delimited_str = ", ".join(values_with_units)
        else:
            values = [f"{value}" for value in self._values]
            comma_delimited_str = ", ".join(values)

        return f"[{comma_delimited_str}]"

    def _create_value_mismatch_exception(self, value: object) -> TypeError:
        return TypeError(
            f"Input type does not match existing type. Input Type: {type(value)} "
            f"Existing Type: {self._value_type}"
        )



"""Waveform timing data types for NI Python APIs."""

from nitypes.waveform._timing._sample_interval import SampleIntervalMode
from nitypes.waveform._timing._timing import Timing

__all__ = [
    "SampleIntervalMode",
    "Timing",
]



from __future__ import annotations

from typing_extensions import Literal

from nitypes.waveform.errors import (
    CapacityMismatchError,
    CapacityTooSmallError,
    DatatypeMismatchError,
    IrregularTimestampCountMismatchError,
    StartIndexTooLargeError,
    StartIndexOrSampleCountTooLargeError,
    NoTimestampInformationError,
    SampleIntervalModeMismatchError,
    SignalCountMismatchError,
)


def create_capacity_mismatch_error(capacity: int, array_length: int) -> CapacityMismatchError:
    """Create an error for a capacity-length mismatch."""
    message = (
        f"The capacity must match the input array length.\n\n"
        f"Capacity: {capacity}\n"
        f"Array length: {array_length}"
    )
    return CapacityMismatchError(message)


def create_capacity_too_small_error(
    capacity: int, min_capacity: int, object_description: str
) -> CapacityTooSmallError:
    """Create an error for when capacity is too small."""
    message = (
        f"The capacity must be equal to or greater than the number of samples in the {object_description}.\n\n"
        f"Capacity: {capacity}\n"
        f"Number of samples: {min_capacity}"
    )
    return CapacityTooSmallError(message)


def create_datatype_mismatch_error(
    arg_description: Literal["input array", "input spectrum", "input waveform"],
    arg_dtype: object,
    other_description: Literal["requested", "spectrum", "waveform"],
    other_dtype: object,
) -> DatatypeMismatchError:
    """Create an error for a data type mismatch."""
    arg_key = {
        "input array": "Input array data type",
        "input spectrum": "Input spectrum data type",
        "input waveform": "Input waveform data type",
    }
    other_key = {
        "requested": "Requested data type",
        "spectrum": "Spectrum data type",
        "waveform": "Waveform data type",
    }
    message = (
        f"The data type of the {arg_description} must match the {other_description} data type.\n\n"
        f"{arg_key[arg_description]}: {arg_dtype}\n"
        f"{other_key[other_description]}: {other_dtype}"
    )
    return DatatypeMismatchError(message)


def create_irregular_timestamp_count_mismatch_error(
    irregular_timestamp_count: int,
    other_description: Literal["input array length", "number of samples in the waveform"],
    other: int,
    *,
    reversed: bool = False,
) -> IrregularTimestampCountMismatchError:
    """Create an error for an irregular timestamp count mismatch."""
    other_key = {
        "input array length": "Array length",
        "number of samples in the waveform": "Number of samples",
    }
    if reversed:
        message = (
            "The input array length must be equal to the number of irregular timestamps.\n\n"
            f"{other_key[other_description]}: {other}\n"
            f"Number of timestamps: {irregular_timestamp_count}"
        )
    else:
        message = (
            f"The number of irregular timestamps must be equal to the {other_description}.\n\n"
            f"Number of timestamps: {irregular_timestamp_count}\n"
            f"{other_key[other_description]}: {other}"
        )
    return IrregularTimestampCountMismatchError(message)


def create_start_index_too_large_error(
    start_index: int,
    capacity_description: Literal[
        "capacity",
        "input array length",
        "number of samples in the spectrum",
        "number of samples in the waveform",
    ],
    capacity: int,
) -> StartIndexTooLargeError:
    """Create an error for an invalid start index argument."""
    capacity_key = {
        "capacity": "Capacity",
        "input array length": "Array length",
        "number of samples in the spectrum": "Number of samples",
        "number of samples in the waveform": "Number of samples",
    }
    message = (
        f"The start index must be less than or equal to the {capacity_description}.\n\n"
        f"Start index: {start_index}\n"
        f"{capacity_key[capacity_description]}: {capacity}"
    )
    return StartIndexTooLargeError(message)


def create_start_index_or_sample_count_too_large_error(
    start_index: int,
    sample_count: int,
    capacity_description: Literal[
        "capacity",
        "input array length",
        "number of samples in the expected waveform",
        "number of samples in the spectrum",
        "number of samples in the waveform",
    ],
    capacity: int,
) -> StartIndexOrSampleCountTooLargeError:
    """Create an error for an invalid start index or sample count argument."""
    capacity_key = {
        "capacity": "Capacity",
        "input array length": "Array length",
        "number of samples in the expected waveform": "Number of samples",
        "number of samples in the spectrum": "Number of samples",
        "number of samples in the waveform": "Number of samples",
    }
    message = (
        f"The sum of the start index and sample count must be less than or equal to the {capacity_description}.\n\n"
        f"Start index: {start_index}\n"
        f"Sample count: {sample_count}\n"
        f"{capacity_key[capacity_description]}: {capacity}"
    )
    return StartIndexOrSampleCountTooLargeError(message)


def create_no_timestamp_information_error() -> NoTimestampInformationError:
    """Create an error for waveform timing with no timestamp information."""
    message = (
        "The waveform timing does not have valid timestamp information. "
        "To obtain timestamps, the waveform must be irregular or must be initialized "
        "with a valid time stamp and sample interval."
    )
    return NoTimestampInformationError(message)


def create_sample_interval_mode_mismatch_error() -> SampleIntervalModeMismatchError:
    """Create an error for mixing none/regular with irregular timing."""
    message = (
        "The timing of one or more waveforms does not match the timing of the current waveform."
    )
    return SampleIntervalModeMismatchError(message)


def create_signal_count_mismatch_error(
    arg_description: Literal["expected waveform", "input array", "input waveform", "provided"],
    arg_signal_count: int,
    other_description: Literal["array", "port", "waveform"],
    other_signal_count: int,
) -> SignalCountMismatchError:
    """Create an error for a mismatched signal count."""
    arg_key = {
        "expected waveform": "Expected waveform signal count",
        "input array": "Input array signal count",
        "input waveform": "Input waveform signal count",
        "provided": "Signal count",
    }
    other_key = {
        "array": "Array signal count",
        "port": "Port signal count",
        "waveform": "Waveform signal count",
    }
    message = (
        f"The {arg_description} signal count must match the {other_description} signal count.\n\n"
        f"{arg_key[arg_description]}: {arg_signal_count}\n"
        f"{other_key[other_description]}: {other_signal_count}"
    )
    return SignalCountMismatchError(message)



from __future__ import annotations

import re


def parse_version(version: str) -> tuple[int, int, int]:
    """Parse a version number into a major/minor/update tuple, ignoring suffixes.

    This is a minimal version parser to avoid requiring the ``packaging`` package.

    >>> parse_version("1.2.3")
    (1, 2, 3)
    >>> parse_version("123.456.789")
    (123, 456, 789)
    >>> parse_version("1.0")
    (1, 0, 0)
    >>> parse_version("1")
    (1, 0, 0)
    >>> parse_version("1.2.3-dev0")
    (1, 2, 3)
    """
    match = re.match(r"^(\d+)(?:\.(\d+)(?:\.(\d+))?)?", version)
    if match is None:
        raise ValueError(f"Invalid version number: {version}")
    return int(match.group(1)), int(match.group(2) or "0"), int(match.group(3) or "0")



from __future__ import annotations

import sys
from collections.abc import Mapping, Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    SupportsFloat,
    SupportsIndex,
    Union,
    final,
    overload,
)

import numpy as np
import numpy.typing as npt
from typing_extensions import Self, TypeVar

from nitypes._arguments import arg_to_float, arg_to_uint, validate_dtype
from nitypes._exceptions import invalid_arg_type, invalid_array_ndim
from nitypes._numpy import asarray as _np_asarray, long as _np_long, ulong as _np_ulong
from nitypes.waveform._exceptions import (
    create_capacity_mismatch_error,
    create_capacity_too_small_error,
    create_datatype_mismatch_error,
    create_start_index_or_sample_count_too_large_error,
    create_start_index_too_large_error,
)
from nitypes.waveform._extended_properties import CHANNEL_NAME, UNIT_DESCRIPTION
from nitypes.waveform.typing import ExtendedPropertyValue

if sys.version_info < (3, 10):
    import array as std_array

if TYPE_CHECKING:
    # Import from the public package so the docs don't reference private submodules.
    from nitypes.waveform import ExtendedPropertyDictionary
else:
    from nitypes.waveform._extended_properties import ExtendedPropertyDictionary

_TData = TypeVar("_TData", bound=Union[np.floating, np.integer])
_TOtherData = TypeVar("_TOtherData", bound=Union[np.floating, np.integer])

# Use the C types here because np.isdtype() considers some of them to be distinct types, even when
# they have the same size (e.g. np.intc vs. np.int_).
_DATA_DTYPES = (
    # Floating point
    np.single,
    np.double,
    # Signed integers
    np.byte,
    np.short,
    np.intc,
    np.int_,
    _np_long,
    np.longlong,
    # Unsigned integers
    np.ubyte,
    np.ushort,
    np.uintc,
    np.uint,
    _np_ulong,
    np.ulonglong,
)


@final
class Spectrum(Generic[_TData]):
    """A frequency spectrum, which encapsulates analog data and frequency information.

    Constructing
    ^^^^^^^^^^^^

    To construct a frequency spectrum, use the :class:`Spectrum` class:

    >>> Spectrum()
    nitypes.waveform.Spectrum(0)
    >>> Spectrum(5)
    nitypes.waveform.Spectrum(5, data=array([0., 0., 0., 0., 0.]))

    To construct a frequency spectrum from a NumPy array, use the :any:`Spectrum.from_array_1d`
    method.

    >>> import numpy as np
    >>> Spectrum.from_array_1d(np.array([1.0, 2.0, 3.0]))
    nitypes.waveform.Spectrum(3, data=array([1., 2., 3.]))

    You can also use :any:`Spectrum.from_array_1d` to construct a frequency spectrum from a
    sequence, such as a list. In this case, you must specify the NumPy data type.

    >>> Spectrum.from_array_1d([1.0, 2.0, 3.0], np.float64)
    nitypes.waveform.Spectrum(3, data=array([1., 2., 3.]))

    The 2D version, :any:`Spectrum.from_array_2d`, returns multiple waveforms, one for each row of
    data in the array or nested sequence.

    >>> nested_list = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
    >>> Spectrum.from_array_2d(nested_list, np.float64)  # doctest: +NORMALIZE_WHITESPACE
    [nitypes.waveform.Spectrum(3, data=array([1., 2., 3.])),
    nitypes.waveform.Spectrum(3, data=array([4., 5., 6.]))]

    Class members
    ^^^^^^^^^^^^^
    """

    @overload
    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[_TOtherData],
        dtype: None = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
    ) -> Spectrum[_TOtherData]: ...

    @overload
    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: type[_TOtherData] | np.dtype[_TOtherData],
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
    ) -> Spectrum[_TOtherData]: ...

    @overload
    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: npt.DTypeLike = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
    ) -> Spectrum[Any]: ...

    @classmethod
    def from_array_1d(
        cls,
        array: npt.NDArray[Any] | Sequence[Any],
        dtype: npt.DTypeLike = None,
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        start_frequency: SupportsFloat | None = None,
        frequency_increment: SupportsFloat | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
    ) -> Spectrum[Any]:
        """Construct a spectrum from a one-dimensional array or sequence.

        Args:
            array: The spectrum data as a one-dimensional array or a sequence.
            dtype: The NumPy data type for the spectrum data. This argument is required
                when array is a sequence.
            copy: Specifies whether to copy the array or save a reference to it.
            start_index: The sample index at which the spectrum data begins.
            sample_count: The number of samples in the spectrum.
            start_frequency: The start frequency of the spectrum.
            frequency_increment: The frequency increment of the spectrum.
            extended_properties: The extended properties of the spectrum.

        Returns:
            A spectrum containing the specified data.
        """
        if isinstance(array, np.ndarray):
            if array.ndim != 1:
                raise invalid_array_ndim(
                    "input array", "one-dimensional array or sequence", array.ndim
                )
        elif isinstance(array, Sequence) or (
            sys.version_info < (3, 10) and isinstance(array, std_array.array)
        ):
            if dtype is None:
                raise ValueError("You must specify a dtype when the input array is a sequence.")
        else:
            raise invalid_arg_type("input array", "one-dimensional array or sequence", array)

        return cls(
            data=_np_asarray(array, dtype, copy=copy),
            start_index=start_index,
            sample_count=sample_count,
            start_frequency=start_frequency,
            frequency_increment=frequency_increment,
            extended_properties=extended_properties,
        )

    @overload
    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[_TOtherData],
        dtype: None = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
    ) -> Sequence[Spectrum[_TOtherData]]: ...

    @overload
    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[Any] | Sequence[Sequence[Any]],
        dtype: type[_TOtherData] | np.dtype[_TOtherData],
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
    ) -> Sequence[Spectrum[_TOtherData]]: ...

    @overload
    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[Any] | Sequence[Sequence[Any]],
        dtype: npt.DTypeLike = ...,
        *,
        copy: bool = ...,
        start_index: SupportsIndex | None = ...,
        sample_count: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
    ) -> Sequence[Spectrum[Any]]: ...

    @classmethod
    def from_array_2d(
        cls,
        array: npt.NDArray[Any] | Sequence[Sequence[Any]],
        dtype: npt.DTypeLike = None,
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
        start_frequency: SupportsFloat | None = None,
        frequency_increment: SupportsFloat | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
    ) -> Sequence[Spectrum[Any]]:
        """Construct a list of spectrums from a two-dimensional array or nested sequence.

        Args:
            array: The spectrum data as a two-dimensional array or a nested sequence.
            dtype: The NumPy data type for the spectrum data. This argument is required
                when array is a sequence.
            copy: Specifies whether to copy the array or save a reference to it.
            start_index: The sample index at which the spectrum data begins.
            sample_count: The number of samples in the spectrum.
            start_frequency: The start frequency of the spectrum.
            frequency_increment: The frequency increment of the spectrum.
            extended_properties: The extended properties of the spectrum.

        Returns:
            A list containing a spectrum for each row of the specified data.

        When constructing multiple spectrums, the same extended properties, timing
        information, and scale mode are applied to all spectrums. Consider assigning
        these properties after construction.
        """
        if isinstance(array, np.ndarray):
            if array.ndim != 2:
                raise invalid_array_ndim(
                    "input array", "two-dimensional array or nested sequence", array.ndim
                )
        elif isinstance(array, Sequence) or (
            sys.version_info < (3, 10) and isinstance(array, std_array.array)
        ):
            if dtype is None:
                raise ValueError("You must specify a dtype when the input array is a sequence.")
        else:
            raise invalid_arg_type("input array", "two-dimensional array or nested sequence", array)

        return [
            cls(
                data=_np_asarray(array[i], dtype, copy=copy),
                start_index=start_index,
                sample_count=sample_count,
                start_frequency=start_frequency,
                frequency_increment=frequency_increment,
                extended_properties=extended_properties,
            )
            for i in range(len(array))
        ]

    __slots__ = [
        "_data",
        "_start_index",
        "_sample_count",
        "_start_frequency",
        "_frequency_increment",
        "_extended_properties",
        "__weakref__",
    ]

    _data: npt.NDArray[_TData]
    _start_index: int
    _sample_count: int
    _start_frequency: float
    _frequency_increment: float
    _extended_properties: ExtendedPropertyDictionary

    # If neither dtype nor data is specified, _TData defaults to np.float64.
    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: Spectrum[np.float64],
        sample_count: SupportsIndex | None = ...,
        dtype: None = ...,
        *,
        data: None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: Spectrum[_TOtherData],
        sample_count: SupportsIndex | None = ...,
        dtype: type[_TOtherData] | np.dtype[_TOtherData] = ...,
        *,
        data: None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: Spectrum[_TOtherData],
        sample_count: SupportsIndex | None = ...,
        dtype: None = ...,
        *,
        data: npt.NDArray[_TOtherData] = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
    ) -> None: ...

    @overload
    def __init__(  # noqa: D107 - Missing docstring in __init__ (auto-generated noqa)
        self: Spectrum[Any],
        sample_count: SupportsIndex | None = ...,
        dtype: npt.DTypeLike = ...,
        *,
        data: npt.NDArray[Any] | None = ...,
        start_index: SupportsIndex | None = ...,
        capacity: SupportsIndex | None = ...,
        start_frequency: SupportsFloat | None = ...,
        frequency_increment: SupportsFloat | None = ...,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = ...,
        copy_extended_properties: bool = ...,
    ) -> None: ...

    def __init__(
        self,
        sample_count: SupportsIndex | None = None,
        dtype: npt.DTypeLike = None,
        *,
        data: npt.NDArray[Any] | None = None,
        start_index: SupportsIndex | None = None,
        capacity: SupportsIndex | None = None,
        start_frequency: SupportsFloat | None = None,
        frequency_increment: SupportsFloat | None = None,
        extended_properties: Mapping[str, ExtendedPropertyValue] | None = None,
        copy_extended_properties: bool = True,
    ) -> None:
        """Initialize a new frequency spectrum.

        Args:
            sample_count: The number of samples in the spectrum.
            dtype: The NumPy data type for the spectrum data.
            data: A NumPy ndarray to use for sample storage. The spectrum takes ownership
                of this array. If not specified, an ndarray is created based on the specified dtype,
                start index, sample count, and capacity.
            start_index: The sample index at which the spectrum data begins.
            sample_count: The number of samples in the spectrum.
            capacity: The number of samples to allocate. Pre-allocating a larger buffer optimizes
                appending samples to the spectrum.
            start_frequency: The start frequency of the spectrum.
            frequency_increment: The frequency increment of the spectrum.
            extended_properties: The extended properties of the spectrum.
            copy_extended_properties: Specifies whether to copy the extended properties or take
                ownership.

        Returns:
            A frequency spectrum.
        """
        if data is None:
            self._init_with_new_array(
                sample_count, dtype, start_index=start_index, capacity=capacity
            )
        elif isinstance(data, np.ndarray):
            self._init_with_provided_array(
                data,
                dtype,
                start_index=start_index,
                sample_count=sample_count,
                capacity=capacity,
            )
        else:
            raise invalid_arg_type("raw data", "NumPy ndarray", data)

        self._start_frequency = arg_to_float("start frequency", start_frequency, 0.0)
        self._frequency_increment = arg_to_float("frequency increment", frequency_increment, 0.0)

        if copy_extended_properties or not isinstance(
            extended_properties, ExtendedPropertyDictionary
        ):
            extended_properties = ExtendedPropertyDictionary(extended_properties)
        self._extended_properties = extended_properties

    def _init_with_new_array(
        self,
        sample_count: SupportsIndex | None = None,
        dtype: npt.DTypeLike = None,
        *,
        start_index: SupportsIndex | None = None,
        capacity: SupportsIndex | None = None,
    ) -> None:
        start_index = arg_to_uint("start index", start_index, 0)
        sample_count = arg_to_uint("sample count", sample_count, 0)
        capacity = arg_to_uint("capacity", capacity, sample_count)

        if dtype is None:
            dtype = np.float64
        validate_dtype(dtype, _DATA_DTYPES)

        if start_index > capacity:
            raise create_start_index_too_large_error(start_index, "capacity", capacity)
        if start_index + sample_count > capacity:
            raise create_start_index_or_sample_count_too_large_error(
                start_index, sample_count, "capacity", capacity
            )

        self._data = np.zeros(capacity, dtype)
        self._start_index = start_index
        self._sample_count = sample_count

    def _init_with_provided_array(
        self,
        data: npt.NDArray[_TData],
        dtype: npt.DTypeLike = None,
        *,
        start_index: SupportsIndex | None = None,
        sample_count: SupportsIndex | None = None,
        capacity: SupportsIndex | None = None,
    ) -> None:
        if not isinstance(data, np.ndarray):
            raise invalid_arg_type("input array", "one-dimensional array", data)
        if data.ndim != 1:
            raise invalid_array_ndim("input array", "one-dimensional array", data.ndim)

        if dtype is None:
            dtype = data.dtype
        if dtype != data.dtype:
            raise create_datatype_mismatch_error(
                "input array", data.dtype, "requested", np.dtype(dtype)
            )
        validate_dtype(dtype, _DATA_DTYPES)

        capacity = arg_to_uint("capacity", capacity, len(data))
        if capacity != len(data):
            raise create_capacity_mismatch_error(capacity, len(data))

        start_index = arg_to_uint("start index", start_index, 0)
        if start_index > capacity:
            raise create_start_index_too_large_error(start_index, "input array length", capacity)

        sample_count = arg_to_uint("sample count", sample_count, len(data) - start_index)
        if start_index + sample_count > len(data):
            raise create_start_index_or_sample_count_too_large_error(
                start_index, sample_count, "input array length", len(data)
            )

        self._data = data
        self._start_index = start_index
        self._sample_count = sample_count

    @property
    def data(self) -> npt.NDArray[_TData]:
        """The spectrum data."""
        return self._data[self._start_index : self._start_index + self._sample_count]

    def get_data(
        self, start_index: SupportsIndex | None = 0, sample_count: SupportsIndex | None = None
    ) -> npt.NDArray[_TData]:
        """Get a subset of the spectrum data.

        Args:
            start_index: The sample index at which the data begins.
            sample_count: The number of samples to return.

        Returns:
            A subset of the spectrum data.
        """
        start_index = arg_to_uint("start index", start_index, 0)
        if start_index > self.sample_count:
            raise create_start_index_too_large_error(
                start_index, "number of samples in the spectrum", self.sample_count
            )

        sample_count = arg_to_uint("sample count", sample_count, self.sample_count - start_index)
        if start_index + sample_count > self.sample_count:
            raise create_start_index_or_sample_count_too_large_error(
                start_index, sample_count, "number of samples in the spectrum", self.sample_count
            )

        return self.data[start_index : start_index + sample_count]

    @property
    def sample_count(self) -> int:
        """The number of samples in the spectrum."""
        return self._sample_count

    @property
    def capacity(self) -> int:
        """The total capacity available for spectrum data.

        Setting the capacity resizes the underlying NumPy array in-place.

        * Other Python objects with references to the array will see the array size change.
        * If the array has a reference to an external buffer (such as an array.array), attempting
          to resize it raises ValueError.
        """
        return len(self._data)

    @capacity.setter
    def capacity(self, value: int) -> None:
        value = arg_to_uint("capacity", value)
        min_capacity = self._start_index + self._sample_count
        if value < min_capacity:
            raise create_capacity_too_small_error(value, min_capacity, "spectrum")
        if value != len(self._data):
            self._data.resize(value, refcheck=False)

    @property
    def dtype(self) -> np.dtype[_TData]:
        """The NumPy dtype for the spectrum data."""
        return self._data.dtype

    @property
    def start_frequency(self) -> float:
        """The start frequency of the spectrum."""
        return self._start_frequency

    @start_frequency.setter
    def start_frequency(self, value: float) -> None:
        if not isinstance(value, (float, int)):
            raise invalid_arg_type("start frequency", "float", value)
        self._start_frequency = value

    @property
    def frequency_increment(self) -> float:
        """The frequency increment of the spectrum."""
        return self._frequency_increment

    @frequency_increment.setter
    def frequency_increment(self, value: float) -> None:
        if not isinstance(value, (float, int)):
            raise invalid_arg_type("frequency increment", "float", value)
        self._frequency_increment = value

    @property
    def extended_properties(self) -> ExtendedPropertyDictionary:
        """The extended properties for the spectrum."""
        return self._extended_properties

    @property
    def channel_name(self) -> str:
        """The name of the device channel from which the spectrum was acquired."""
        value = self._extended_properties.get(CHANNEL_NAME, "")
        assert isinstance(value, str)
        return value

    @channel_name.setter
    def channel_name(self, value: str) -> None:
        if not isinstance(value, str):
            raise invalid_arg_type("channel name", "str", value)
        self._extended_properties[CHANNEL_NAME] = value

    @property
    def unit_description(self) -> str:
        """The unit of measurement, such as volts, of the spectrum."""
        value = self._extended_properties.get(UNIT_DESCRIPTION, "")
        assert isinstance(value, str)
        return value

    @unit_description.setter
    def unit_description(self, value: str) -> None:
        if not isinstance(value, str):
            raise invalid_arg_type("unit description", "str", value)
        self._extended_properties[UNIT_DESCRIPTION] = value

    def append(
        self,
        other: npt.NDArray[_TData] | Spectrum[_TData] | Sequence[Spectrum[_TData]],
        /,
    ) -> None:
        """Append data to the spectrum.

        Args:
            other: The array or spectrum(s) to append.

        Raises:
            ValueError: The other array has the wrong number of dimensions.
            TypeError: The data types of the current spectrum and other array or spectrum(s) do not
                match, or an argument has the wrong data type.

        When appending spectrums:

        * Extended properties of the other spectrum(s) are merged into the current spectrum if they
          are not already set in the current spectrum.
        """
        if isinstance(other, np.ndarray):
            self._append_array(other)
        elif isinstance(other, Spectrum):
            self._append_spectrum(other)
        elif isinstance(other, Sequence) and all(isinstance(x, Spectrum) for x in other):
            self._append_spectrums(other)
        else:
            raise invalid_arg_type("input", "array or spectrum(s)", other)

    def _append_array(
        self,
        array: npt.NDArray[_TData],
    ) -> None:
        if array.dtype != self.dtype:
            raise create_datatype_mismatch_error("input array", array.dtype, "spectrum", self.dtype)
        if array.ndim != 1:
            raise invalid_array_ndim("input array", "one-dimensional array", array.ndim)

        self._increase_capacity(len(array))

        offset = self._start_index + self._sample_count
        self._data[offset : offset + len(array)] = array
        self._sample_count += len(array)

    def _append_spectrum(self, spectrum: Spectrum[_TData]) -> None:
        self._append_spectrums([spectrum])

    def _append_spectrums(self, spectrums: Sequence[Spectrum[_TData]]) -> None:
        for spectrum in spectrums:
            if spectrum.dtype != self.dtype:
                raise create_datatype_mismatch_error(
                    "input spectrum", spectrum.dtype, "spectrum", self.dtype
                )

        self._increase_capacity(sum(spectrum.sample_count for spectrum in spectrums))

        offset = self._start_index + self._sample_count
        for spectrum in spectrums:
            self._data[offset : offset + spectrum.sample_count] = spectrum.data
            offset += spectrum.sample_count
            self._sample_count += spectrum.sample_count
            self._extended_properties._merge(spectrum._extended_properties)

    def _increase_capacity(self, amount: int) -> None:
        new_capacity = self._start_index + self._sample_count + amount
        if new_capacity > self.capacity:
            self.capacity = new_capacity

    def load_data(
        self,
        array: npt.NDArray[_TData],
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
    ) -> None:
        """Load new data into an existing spectrum.

        Args:
            array: A NumPy array containing the data to load.
            copy: Specifies whether to copy the array or save a reference to it.
            start_index: The sample index at which the spectrum data begins.
            sample_count: The number of samples in the spectrum.
        """
        if isinstance(array, np.ndarray):
            self._load_array(array, copy=copy, start_index=start_index, sample_count=sample_count)
        else:
            raise invalid_arg_type("input array", "array", array)

    def _load_array(
        self,
        array: npt.NDArray[_TData],
        *,
        copy: bool = True,
        start_index: SupportsIndex | None = 0,
        sample_count: SupportsIndex | None = None,
    ) -> None:
        if array.dtype != self.dtype:
            raise create_datatype_mismatch_error("input array", array.dtype, "spectrum", self.dtype)
        if array.ndim != 1:
            raise invalid_array_ndim("input array", "one-dimensional array", array.ndim)

        start_index = arg_to_uint("start index", start_index, 0)
        sample_count = arg_to_uint("sample count", sample_count, len(array) - start_index)

        if copy:
            if sample_count > len(self._data):
                self.capacity = sample_count
            self._data[0:sample_count] = array[start_index : start_index + sample_count]
            self._start_index = 0
            self._sample_count = sample_count
        else:
            self._data = array
            self._start_index = start_index
            self._sample_count = sample_count

    def __eq__(self, value: object, /) -> bool:
        """Return self==value."""
        if not isinstance(value, self.__class__):
            return NotImplemented
        return (
            self.dtype == value.dtype
            and np.array_equal(self.data, value.data)
            and self.start_frequency == value.start_frequency
            and self.frequency_increment == value.frequency_increment
            and self._extended_properties == value._extended_properties
        )

    def __reduce__(self) -> tuple[Any, ...]:
        """Return object state for pickling."""
        ctor_args = (self._sample_count, self.dtype)
        ctor_kwargs: dict[str, Any] = {
            "data": self.data,
            "start_frequency": self._start_frequency,
            "frequency_increment": self._frequency_increment,
            "extended_properties": self._extended_properties,
            "copy_extended_properties": False,
        }
        return (self.__class__._unpickle, (ctor_args, ctor_kwargs))

    @classmethod
    def _unpickle(cls, args: tuple[Any, ...], kwargs: dict[str, Any]) -> Self:
        return cls(*args, **kwargs)

    def __repr__(self) -> str:
        """Return repr(self)."""
        args = [f"{self._sample_count}"]
        if self.dtype != np.float64:
            args.append(f"{self.dtype.name}")
        # start_index and capacity are not shown because they are allocation details. data hides
        # the unused data before start_index and after start_index+sample_count.
        if self._sample_count > 0:
            args.append(f"data={self.data!r}")
        if self._start_frequency != 0.0:
            args.append(f"start_frequency={self.start_frequency!r}")
        if self._frequency_increment != 0.0:
            args.append(f"frequency_increment={self._frequency_increment!r}")
        if self._extended_properties:
            args.append(f"extended_properties={self._extended_properties._properties!r}")
        return f"{self.__class__.__module__}.{self.__class__.__name__}({', '.join(args)})"



from __future__ import annotations

from typing import NamedTuple


class TimeValueTuple(NamedTuple):
    """A named tuple containing the whole seconds and fractional seconds parts of a time value."""

    whole_seconds: int
    """The whole seconds portion of a binary time value. This should be an int64."""

    fractional_seconds: int
    """The fractional seconds portion of a binary time value. This should be a uint64."""

    @staticmethod
    def from_cvi(lsb: int, msb: int) -> TimeValueTuple:
        """Create a :class:`TimeValueTuple` from a ``CVIAbsoluteTime`` representation."""
        return TimeValueTuple(whole_seconds=msb, fractional_seconds=lsb)

    def to_cvi(self) -> tuple[int, int]:
        """Return a representation as ``CVIAbsoluteTime``."""
        return (self.fractional_seconds, self.whole_seconds)
